"""This packages includes CostFunction and generator implementations."""
from __future__ import annotations

from bqskit.ir.opt.cost.functions.cost import HilbertSchmidtCost
from bqskit.ir.opt.cost.functions.cost import HilbertSchmidtCostGenerator
from bqskit.ir.opt.cost.functions.residuals import HilbertSchmidtResiduals
from bqskit.ir.opt.cost.functions.residuals import (
    HilbertSchmidtResidualsGenerator,
)

__all__ = [
    'HilbertSchmidtCost',
    'HilbertSchmidtCostGenerator',
    'HilbertSchmidtResiduals',
    'HilbertSchmidtResidualsGenerator',
]



"""This module implements a OPENQASM parse tree visitor."""
from __future__ import annotations

import logging
import os
from dataclasses import dataclass
from typing import Any
from typing import cast
from typing import List
from typing import NamedTuple
from typing import Tuple

import lark
import numpy as np
from lark import Visitor

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.composed.controlled import ControlledGate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.constant.b import BGate
from bqskit.ir.gates.constant.ccx import CCXGate
from bqskit.ir.gates.constant.ch import CHGate
from bqskit.ir.gates.constant.cs import CSGate
from bqskit.ir.gates.constant.ct import CTGate
from bqskit.ir.gates.constant.cx import CXGate
from bqskit.ir.gates.constant.cy import CYGate
from bqskit.ir.gates.constant.cz import CZGate
from bqskit.ir.gates.constant.ecr import ECRGate
from bqskit.ir.gates.constant.h import HGate
from bqskit.ir.gates.constant.identity import IdentityGate
from bqskit.ir.gates.constant.iswap import ISwapGate
from bqskit.ir.gates.constant.itoffoli import IToffoliGate
from bqskit.ir.gates.constant.rccx import RC3XGate
from bqskit.ir.gates.constant.rccx import RCCXGate
from bqskit.ir.gates.constant.s import SGate
from bqskit.ir.gates.constant.sdg import SdgGate
from bqskit.ir.gates.constant.sqrtcnot import SqrtCNOTGate
from bqskit.ir.gates.constant.sqrtiswap import SqrtISwapGate
from bqskit.ir.gates.constant.swap import SwapGate
from bqskit.ir.gates.constant.sx import SXGate
from bqskit.ir.gates.constant.sycamore import SycamoreGate
from bqskit.ir.gates.constant.t import TGate
from bqskit.ir.gates.constant.tdg import TdgGate
from bqskit.ir.gates.constant.x import XGate
from bqskit.ir.gates.constant.xx import XXGate
from bqskit.ir.gates.constant.y import YGate
from bqskit.ir.gates.constant.yy import YYGate
from bqskit.ir.gates.constant.z import ZGate
from bqskit.ir.gates.constant.zz import ZZGate
from bqskit.ir.gates.measure import MeasurementPlaceholder
from bqskit.ir.gates.parameterized.ccp import CCPGate
from bqskit.ir.gates.parameterized.cp import CPGate
from bqskit.ir.gates.parameterized.crx import CRXGate
from bqskit.ir.gates.parameterized.cry import CRYGate
from bqskit.ir.gates.parameterized.crz import CRZGate
from bqskit.ir.gates.parameterized.cu import CUGate
from bqskit.ir.gates.parameterized.fsim import FSIMGate
from bqskit.ir.gates.parameterized.phasedxz import PhasedXZGate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rxx import RXXGate
from bqskit.ir.gates.parameterized.ry import RYGate
from bqskit.ir.gates.parameterized.ryy import RYYGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.rzz import RZZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.ir.gates.parameterized.u1q import U1qGate
from bqskit.ir.gates.parameterized.u2 import U2Gate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.gates.reset import Reset
from bqskit.ir.lang.language import LangException
from bqskit.ir.lang.qasm2.parser import parse
from bqskit.ir.location import CircuitLocation
from bqskit.ir.operation import Operation


_logger = logging.getLogger(__name__)


class QubitReg(NamedTuple):
    """Definition of a Qubit Register."""
    name: str
    size: int


class ClassicalReg(NamedTuple):
    """Definition of a Classical Bit Register."""
    name: str
    size: int


@dataclass
class GateDef:
    """Definition of a QASM Gate."""
    qasm_name: str
    num_params: int
    num_vars: int
    gate: Gate

    def build_op(self, loc: CircuitLocation, params: list[float]) -> Operation:
        """Build an operation for this gate."""
        return Operation(self.gate, loc, params)


@dataclass
class CustomGateDef:
    """Definition of a Custom QASM Gate."""
    qasm_name: str
    num_params: int
    num_vars: int
    gate_def_list: list[GateDef | CustomGateDef]
    loc_list: list[CircuitLocation]
    param_exp_list: list[list[lark.Tree | float]]

    def build_op(self, loc: CircuitLocation, params: list[float]) -> Operation:
        """Build an operation for this gate."""
        cgc = Circuit(self.num_vars)
        for i in range(len(self.gate_def_list)):
            sublocation = self.loc_list[i]
            subparams = self.evaluate_param_exps(self.param_exp_list[i], params)
            cgc.append(self.gate_def_list[i].build_op(sublocation, subparams))
        gate = CircuitGate(cgc, True)
        return Operation(gate, loc, gate._circuit.params)

    def evaluate_param_exps(
        self,
        param_exps: list[lark.Tree | float],
        params: list[float],
    ) -> list[float]:
        """Substitute params into the place holders in param_exp and
        evaluate."""
        eval_params = []
        for exp in param_exps:
            if isinstance(exp, float):
                eval_params.append(exp)
            else:
                subbed_exp = self.replace_param_indices(exp, params)
                eval_params.append(float(eval_exp(subbed_exp)))
        return eval_params

    def replace_param_indices(
        self,
        exp: lark.Tree,
        params: list[float],
    ) -> lark.Tree:
        """Return a new tree with parameter indices replaced with values."""
        if isinstance(exp, lark.Token):
            if exp.type == 'PARAM_IDX':
                return lark.Token('REAL', params[int(exp)])
            else:
                return exp
        children = [self.replace_param_indices(c, params) for c in exp.children]
        return lark.Tree(exp.data, children)


class OPENQASMVisitor(Visitor):
    """
    The OPENQASMVisitor class.

    This class defines a complete visitor for a parse tree produced by
    OPENQASMParser.
    """

    def __init__(self) -> None:
        """Initialize state for the OPENQASMVisitor."""
        self.op_list: list[Operation] = []
        self.qubit_regs: list[QubitReg] = []
        self.gate_defs: dict[str, GateDef] = {}
        self.classical_regs: list[ClassicalReg] = []
        self.gate_def_parsing_obj: Any = None
        self.custom_gate_defs: dict[str, CustomGateDef] = {}
        self.fill_gate_defs()

    def get_circuit(self) -> Circuit:
        """Retrieve the circuit generated after walking a parse tree."""
        num_qubits = sum(qubit_reg.size for qubit_reg in self.qubit_regs)
        if num_qubits == 0:
            raise LangException('No qubit registers defined.')
        circuit = Circuit(num_qubits)
        circuit.extend(self.op_list)

        return circuit

    def fill_gate_defs(self) -> None:
        """Prefills gate definitions with built-in gates."""
        # Parameterized Gates
        self.gate_defs['p'] = GateDef('p', 1, 1, RZGate())
        self.gate_defs['cp'] = GateDef('cp', 1, 2, CPGate())
        self.gate_defs['ccp'] = GateDef('ccp', 1, 3, CCPGate())
        self.gate_defs['crx'] = GateDef('crx', 1, 2, CRXGate())
        self.gate_defs['cry'] = GateDef('cry', 1, 2, CRYGate())
        self.gate_defs['crz'] = GateDef('crz', 1, 2, CRZGate())
        self.gate_defs['fsim'] = GateDef('fsim', 2, 2, FSIMGate())
        self.gate_defs['rx'] = GateDef('rx', 1, 1, RXGate())
        self.gate_defs['pxz'] = GateDef('pxz', 1, 3, PhasedXZGate())
        self.gate_defs['rxx'] = GateDef('rxx', 1, 2, RXXGate())
        self.gate_defs['ry'] = GateDef('ry', 1, 1, RYGate())
        self.gate_defs['ryy'] = GateDef('ryy', 1, 2, RYYGate())
        self.gate_defs['rz'] = GateDef('rz', 1, 1, RZGate())
        self.gate_defs['rzz'] = GateDef('rzz', 1, 2, RZZGate())
        self.gate_defs['u1'] = GateDef('u1', 1, 1, U1Gate())
        self.gate_defs['u1q'] = GateDef('u1q', 2, 1, U1qGate())
        self.gate_defs['U1q'] = GateDef('u1q', 2, 1, U1qGate())
        self.gate_defs['u2'] = GateDef('u2', 2, 1, U2Gate())
        self.gate_defs['u3'] = GateDef('u3', 3, 1, U3Gate())
        self.gate_defs['U'] = GateDef('U', 3, 1, U3Gate())
        self.gate_defs['u'] = GateDef('u', 3, 1, U3Gate())
        self.gate_defs['cu1'] = GateDef('cu1', 1, 2, ControlledGate(U1Gate()))
        self.gate_defs['cu2'] = GateDef('cu2', 2, 2, ControlledGate(U2Gate()))
        self.gate_defs['cu3'] = GateDef('cu3', 3, 2, ControlledGate(U3Gate()))
        self.gate_defs['cu'] = GateDef('cu', 4, 2, CUGate())

        # Constant Gates
        self.gate_defs['b'] = GateDef('b', 0, 2, BGate())
        self.gate_defs['ccx'] = GateDef('ccx', 0, 3, CCXGate())
        self.gate_defs['ch'] = GateDef('ch', 0, 2, CHGate())
        self.gate_defs['cs'] = GateDef('cs', 0, 2, CSGate())
        self.gate_defs['ct'] = GateDef('ct', 0, 2, CTGate())
        self.gate_defs['cx'] = GateDef('cx', 0, 2, CXGate())
        self.gate_defs['CX'] = GateDef('CX', 0, 2, CXGate())
        self.gate_defs['cy'] = GateDef('cy', 0, 2, CYGate())
        self.gate_defs['cz'] = GateDef('cz', 0, 2, CZGate())
        self.gate_defs['ecr'] = GateDef('ecr', 0, 2, ECRGate())
        self.gate_defs['h'] = GateDef('h', 0, 1, HGate())
        self.gate_defs['id'] = GateDef('id', 0, 1, IdentityGate(1))
        self.gate_defs['u0'] = GateDef('u0', 0, 1, IdentityGate(1))
        self.gate_defs['iswap'] = GateDef('iswap', 0, 2, ISwapGate())
        self.gate_defs['iccx'] = GateDef('iccx', 0, 3, IToffoliGate())
        self.gate_defs['s'] = GateDef('s', 0, 1, SGate())
        self.gate_defs['sdg'] = GateDef('sdg', 0, 1, SdgGate())
        self.gate_defs['csx'] = GateDef('csx', 0, 2, SqrtCNOTGate())
        self.gate_defs['cv'] = GateDef('cv', 0, 2, SqrtCNOTGate())
        self.gate_defs['sqisw'] = GateDef('sqisw', 0, 2, SqrtISwapGate())
        self.gate_defs['swap'] = GateDef('swap', 0, 2, SwapGate())
        self.gate_defs['cswap'] = GateDef(
            'cswap', 0, 3, ControlledGate(SwapGate()),
        )
        self.gate_defs['sx'] = GateDef('sx', 0, 1, SXGate())
        self.gate_defs['v'] = GateDef('v', 0, 1, SXGate())
        self.gate_defs['syc'] = GateDef('syc', 0, 2, SycamoreGate())
        self.gate_defs['t'] = GateDef('t', 0, 1, TGate())
        self.gate_defs['tdg'] = GateDef('tdg', 0, 1, TdgGate())
        self.gate_defs['x'] = GateDef('x', 0, 1, XGate())
        self.gate_defs['xx'] = GateDef('xx', 0, 2, XXGate())
        self.gate_defs['y'] = GateDef('y', 0, 1, YGate())
        self.gate_defs['yy'] = GateDef('yy', 0, 2, YYGate())
        self.gate_defs['z'] = GateDef('z', 0, 1, ZGate())
        self.gate_defs['zz'] = GateDef('zz', 0, 2, ZZGate())
        self.gate_defs['sxdg'] = GateDef('sxdg', 0, 1, DaggerGate(SXGate()))
        self.gate_defs['c3x'] = GateDef('c3x', 0, 4, ControlledGate(XGate(), 3))
        self.gate_defs['c4x'] = GateDef('c4x', 0, 5, ControlledGate(XGate(), 4))
        self.gate_defs['c3sqrtx'] = GateDef(
            'c3sqrtx', 0, 4, ControlledGate(SXGate(), 3),
        )
        self.gate_defs['rccx'] = GateDef('rccx', 0, 3, RCCXGate())
        self.gate_defs['rc3x'] = GateDef('rc3x', 0, 4, RC3XGate())

    def qreg(self, tree: lark.Tree) -> None:
        """Qubit register node visitor."""
        reg_name = tree.children[0]
        if any(reg_name == reg.name for reg in self.qubit_regs):
            raise LangException('Qubit register redeclared: %s.' % reg_name)
        reg_size = int(tree.children[1])
        reg = QubitReg(reg_name, reg_size)
        _logger.debug('Qubit register %s declared with size %d.' % reg)
        self.qubit_regs.append(reg)

    def barrier(self, tree: lark.Tree) -> None:
        """Apply a barrier to the circuit."""
        qlist = tree.children[-1]
        location = CircuitLocation(self.convert_qubit_ids_to_indices(qlist))
        gate = BarrierPlaceholder(len(location))
        op = Operation(gate, location)
        self.op_list.append(op)

    def barrierp(self, tree: lark.Tree) -> None:
        raise LangException(
            'BQSKit currently does not support barriers in gate declarations.',
        )

    def gate(self, tree: lark.Tree) -> None:
        """Apply a normal gate statement to the circuit."""
        # Parse parameters
        if len(tree.children) == 3:
            exp_list = self.flatten_exps(tree.children[1])
            params = [float(eval_exp(exp)) for exp in exp_list]
        else:
            params = []

        # Parse location
        qlist = tree.children[-1]
        location = CircuitLocation(self.convert_qubit_ids_to_indices(qlist))

        # Parse gate object
        gate_name = str(tree.children[0])
        if gate_name in self.gate_defs:
            gate_def: GateDef | CustomGateDef = self.gate_defs[gate_name]
        elif gate_name in self.custom_gate_defs:
            gate_def = self.custom_gate_defs[gate_name]
        else:
            raise LangException('Unrecognized gate: %s.' % gate_name)

        if len(params) != gate_def.num_params:
            raise LangException(
                'Expected %d params got %d params for gate %s.'
                % (gate_def.num_params, len(params), gate_name),
            )

        if len(location) != gate_def.num_vars:
            raise LangException(
                'Gate acts on %d qubits, got %d qubit variables.'
                % (gate_def.num_vars, len(location)),
            )

        # Build operation and add to circuit
        self.op_list.append(gate_def.build_op(location, params))

    def cxgate(self, tree: lark.Tree) -> None:
        """CX gate node visitor."""
        control = tree.children[0].children
        target = tree.children[1].children
        cname, cidx = str(control[0]), int(control[1])
        tname, tidx = str(target[0]), int(target[1])

        outer_idx = 0
        cloc: int | None = None
        tloc: int | None = None
        for reg in self.qubit_regs:
            if reg.name == cname:
                cloc = outer_idx + cidx
            if reg.name == tname:
                tloc = outer_idx + tidx
            outer_idx += reg.size

        if cloc is None:
            raise LangException(f'Qubit register not found: {cname}')

        if tloc is None:
            raise LangException(f'Qubit register not found: {tname}')

        if cloc == tloc:
            raise LangException('CX control and target qubit cannot be equal.')

        loc = CircuitLocation([cloc, tloc])
        self.op_list.append(Operation(CXGate(), loc))

    def ugate(self, tree: lark.Tree) -> None:
        """U gate node visitor."""
        exp_list = self.flatten_exps(tree.children[0])
        params = [float(eval_exp(exp)) for exp in exp_list]
        qubit = tree.children[1].children
        name, idx = str(qubit[0]), int(qubit[1])

        outer_idx = 0
        location: int | None = None
        for reg in self.qubit_regs:
            if reg.name == name:
                location = outer_idx + idx
                break
            outer_idx += reg.size

        if location is None:
            raise LangException(f'Qubit register not found: {name}')

        location = CircuitLocation(location)
        self.op_list.append(self.gate_defs['U'].build_op(location, params))

    def gatep(self, tree: lark.Tree) -> None:
        """Apply a gate to a currently-being-built gate declaration."""
        # Parse parameters or store their expression
        param_exps = []
        if len(tree.children) == 3:
            exp_list = self.flatten_exps(tree.children[1])
            for exp in exp_list:
                if self.has_param_variable(exp):
                    param_exps.append(self.replace_param_ids(exp))
                else:
                    param_exps.append(float(eval_exp(exp)))

        self.gate_def_parsing_obj['param_exp_list'].append(param_exps)

        # Parse location
        location: list[int] = []
        iter = tree.children[-1].children[0]
        while iter:
            qname = str(iter.children[-1])
            qid = self.gate_def_parsing_obj['qubits'].index(qname)
            location.insert(0, qid)
            iter = None if len(iter.children) != 2 else iter.children[0]

        self.gate_def_parsing_obj['loc_list'].append(location)

        # Parse gate object
        gate_name = str(tree.children[0])
        if gate_name in self.gate_defs:
            gate_def: GateDef | CustomGateDef = self.gate_defs[gate_name]
        elif gate_name in self.custom_gate_defs:
            gate_def = self.custom_gate_defs[gate_name]
        else:
            raise LangException('Unrecognized gate: %s.' % gate_name)

        if len(param_exps) != gate_def.num_params:
            raise LangException(
                'Expected %d params got %d params for gate %s.'
                % (gate_def.num_params, len(param_exps), gate_name),
            )

        if len(location) != gate_def.num_vars:
            raise LangException(
                'Gate acts on %d qubits, got %d qubit variables.'
                % (gate_def.num_vars, len(location)),
            )

        self.gate_def_parsing_obj['gate_def_list'].append(gate_def)

    def cxgatep(self, tree: lark.Tree) -> None:
        """CX gate node visitor (while building custom gate)."""
        control = tree.children[0].children
        target = tree.children[1].children
        cid = self.gate_def_parsing_obj['qubits'].index(str(control[0]))
        tid = self.gate_def_parsing_obj['qubits'].index(str(target[0]))

        if cid == tid:
            raise LangException('CX control and target qubit cannot be equal.')

        loc = CircuitLocation([cid, tid])
        self.gate_def_parsing_obj['param_exp_list'].append([])
        self.gate_def_parsing_obj['loc_list'].append(loc)
        self.gate_def_parsing_obj['gate_def_list'].append(self.gate_defs['CX'])

    def ugatep(self, tree: lark.Tree) -> None:
        """U gate node visitor (while building custom gate)."""
        param_exps = []
        exp_list = self.flatten_exps(tree.children[0])
        for exp in exp_list:
            if self.has_param_variable(exp):
                param_exps.append(self.replace_param_ids(exp))
            else:
                param_exps.append(float(eval_exp(exp)))

        qubit = tree.children[1].children
        qid = self.gate_def_parsing_obj['qubits'].index(str(qubit[0]))

        location = CircuitLocation(qid)
        self.gate_def_parsing_obj['param_exp_list'].append(param_exps)
        self.gate_def_parsing_obj['loc_list'].append(location)
        self.gate_def_parsing_obj['gate_def_list'].append(self.gate_defs['U'])

    def rbracket(self, tree: lark.Tree) -> None:
        """Finish a gate declaration block."""
        qasm_name = self.gate_def_parsing_obj['name']
        num_params = self.gate_def_parsing_obj['num_params']
        num_vars = self.gate_def_parsing_obj['num_qubits']
        gate_def_list = self.gate_def_parsing_obj['gate_def_list']
        loc_list = self.gate_def_parsing_obj['loc_list']
        param_exp_list = self.gate_def_parsing_obj['param_exp_list']
        self.custom_gate_defs[qasm_name] = CustomGateDef(
            qasm_name,
            num_params,
            num_vars,
            gate_def_list,
            loc_list,
            param_exp_list,
        )
        self.gate_def_parsing_obj = None

    def gatedecl(self, tree: lark.Tree) -> None:
        """Start a gate declaration block."""
        # parse gate name
        self.gate_def_parsing_obj = {'name': str(tree.children[0])}

        # parse parameters
        if len(tree.children) == 2:
            self.gate_def_parsing_obj['num_params'] = 0
            self.gate_def_parsing_obj['params'] = []
        else:
            params: list[str] = []
            iter = tree.children[1]
            while iter:
                if isinstance(iter, lark.Tree):
                    if len(iter.children) == 2:
                        params.insert(0, str(iter.children[1]))
                        iter = iter.children[0]
                    else:
                        params.insert(0, str(iter.children[0]))
                        iter = None
                else:
                    params.insert(0, str(iter))
                    iter = None

            self.gate_def_parsing_obj['num_params'] = len(params)
            self.gate_def_parsing_obj['params'] = params

        # parse qubits
        qubits: list[str] = []
        iter = tree.children[-1]
        while iter:
            if isinstance(iter, lark.Tree):
                if len(iter.children) == 2:
                    qubits.insert(0, str(iter.children[1]))
                    iter = iter.children[0]
                else:
                    qubits.insert(0, str(iter.children[0]))
                    iter = None
            else:
                qubits.insert(0, str(iter))
                iter = None

        self.gate_def_parsing_obj['num_qubits'] = len(qubits)
        self.gate_def_parsing_obj['qubits'] = qubits
        self.gate_def_parsing_obj['gate_def_list'] = []
        self.gate_def_parsing_obj['loc_list'] = []
        self.gate_def_parsing_obj['param_exp_list'] = []

    def flatten_exps(self, explist: lark.Tree) -> list[lark.Tree]:
        """Flatten an explist tree into a list of exp subtrees."""
        exp_list: list[lark.Tree] = []
        iter = explist
        while iter:
            exp_list.insert(0, iter.children[-1])
            iter = None if len(iter.children) == 1 else iter.children[0]
        return exp_list

    def has_param_variable(self, exp: lark.Tree) -> bool:
        """Return true is exp is bound to gatedecl param variable."""
        if isinstance(exp, lark.Token):
            return str(exp) in self.gate_def_parsing_obj['params']
        return any(self.has_param_variable(c) for c in exp.children)

    def replace_param_ids(self, exp: lark.Tree) -> lark.Tree:
        """Return a new tree with parameter ids replaces with indices."""
        if isinstance(exp, lark.Token):
            if str(exp) in self.gate_def_parsing_obj['params']:
                index = self.gate_def_parsing_obj['params'].index(str(exp))
                return lark.Token('PARAM_IDX', index)
            else:
                return exp
        children = [self.replace_param_ids(c) for c in exp.children]
        return lark.Tree(exp.data, children)

    def incstmt(self, tree: lark.Tree) -> None:
        """Include statement node."""
        file_name = str(tree.children[0])
        if file_name[0] == "\"":
            file_name = file_name[1:]
        if file_name[-1] == "\"":
            file_name = file_name[:-1]

        if not os.path.isfile(file_name):
            if file_name != 'qelib1.inc':
                # Suppress warnings over qelib1.inc due to high frequency
                _logger.warning(
                    f'Unable to find {file_name} used in a QASM include'
                    ' statement. Ignoring include file and continuing.',
                )
            return

        with open(file_name) as f:
            filedata = f.read()
            if not filedata.startswith('OPENQASM'):
                filedata = 'OPENQASM 2.0;\n' + filedata
            tree = parse(filedata)
            visitor = OPENQASMVisitor()
            visitor.visit_topdown(tree)
            self.op_list.extend(visitor.op_list)
            self.qubit_regs.extend(visitor.qubit_regs)
            self.custom_gate_defs.update(visitor.custom_gate_defs)

    def creg(self, tree: lark.Tree) -> None:
        """Classical bit register node visitor."""
        reg_name = tree.children[0]
        if any(reg_name == reg.name for reg in self.classical_regs):
            raise LangException('Classical register redeclared: %s.' % reg_name)
        reg_size = int(tree.children[1])
        reg = ClassicalReg(reg_name, reg_size)
        _logger.debug('Classical register %s declared with size %d.' % reg)
        self.classical_regs.append(reg)

    def measure(self, tree: lark.Tree) -> None:
        """Measure statement node visitor."""
        params: list[float] = []
        measurements: dict[int, tuple[str, int]] = {}
        qubit_childs = tree.children[0].children
        class_childs = tree.children[1].children
        qubit_reg_name = str(qubit_childs[0])
        class_reg_name = str(class_childs[0])
        cregs = cast(List[Tuple[str, int]], self.classical_regs)
        qlist = tree.children[0]
        location = CircuitLocation(self.convert_qubit_ids_to_indices(qlist))

        if not any(r.name == qubit_reg_name for r in self.qubit_regs):
            raise LangException(
                f'Measuring undefined qubit register: {qubit_reg_name}',
            )

        if not any(r.name == class_reg_name for r in self.classical_regs):
            raise LangException(
                f'Measuring undefined classical register: {class_reg_name}',
            )

        if len(qubit_childs) == 1 and len(class_childs) == 1:  # for measure all
            for name, size in self.qubit_regs:
                if qubit_reg_name == name:
                    qubit_size = size
                    break
            for name, size in self.classical_regs:
                if class_reg_name == name:
                    class_size = size
                    break
            if qubit_size != class_size:
                raise LangException(
                    'Size mismatch between qubit and classical register'
                    f' during measure operation: {qubit_size} != {class_size}.',
                )

            outer_idx = 0
            for name, size in self.qubit_regs:
                if name == qubit_reg_name:
                    break
                outer_idx += size
            for i in range(qubit_size):
                measurements[outer_idx + i] = (class_reg_name, i)
            mph = MeasurementPlaceholder(cregs, measurements)

        elif len(qubit_childs) == 2 and len(class_childs) == 2:
            # measure qubits to clbits
            qubit_index = int(qubit_childs[1])
            class_index = int(class_childs[1])
            measurements[qubit_index] = (class_reg_name, class_index)
            mph = MeasurementPlaceholder(cregs, measurements)

        else:
            raise LangException(
                'Invalid measurement: either a single qubit is being measured '
                'to a full classical register or a qubit register is being '
                'measured to a single classical bit.',
            )
        op = Operation(mph, location, params)
        self.op_list.append(op)

    def reset(self, tree: lark.Tree) -> None:
        """Reset node visitor."""
        params: list[float] = []
        qlist = tree.children[-1]
        if len(qlist.children) == 2:
            location = CircuitLocation(self.convert_qubit_ids_to_indices(qlist))
            op = Operation(Reset(), location, params)
            self.op_list.append(op)
        else:
            locations = [
                CircuitLocation(i)
                for i in range(self.qubit_regs[0][1])
            ]
            for location in locations:
                op = Operation(Reset(), location, params)
                self.op_list.append(op)

    def convert_qubit_ids_to_indices(self, qlist: lark.Tree) -> list[int]:
        if qlist.data == 'anylist':
            # Anylist is either idlist or mixedlist
            return self.convert_qubit_ids_to_indices(qlist.children[0])

        if qlist.data == 'idlist':
            # List of ids, e.g. q, r, but without indices
            ids = []
            tree_iter = qlist
            while len(tree_iter.children) == 2:
                ids.append(str(tree_iter.children[0]))
                tree_iter = tree_iter.children[1]
            ids.append(str(tree_iter.children[0]))

            out_idxs = []
            for qubit_id in ids:
                out_idxs.extend(self.convert_qubit_id_to_indices(qubit_id))

            return out_idxs

        if qlist.data == 'argument':
            # ID | ID "[" NNINTEGER "]"
            qubit_id = qlist.children[0]
            if len(qlist.children) == 2:
                idx = self.convert_qubit_id_to_first_index(qubit_id)
                return [idx + int(qlist.children[1])]

            return self.convert_qubit_id_to_indices(qubit_id)

        if qlist.data == 'mixedlist':
            if isinstance(qlist.children[0], lark.Tree):
                idxs = self.convert_qubit_ids_to_indices(qlist.children[0])

                if len(qlist.children) == 2:
                    qubit_id = qlist.children[1]
                    return idxs + self.convert_qubit_id_to_indices(qubit_id)

                idx = self.convert_qubit_id_to_first_index(qlist.children[1])
                return idxs + [idx + int(qlist.children[2])]

            else:
                idx = self.convert_qubit_id_to_first_index(qlist.children[0])
                return [idx + int(qlist.children[1])]

        raise LangException('Incorrect qlist type.')

    def convert_qubit_id_to_first_index(self, qubit_id: str) -> int:
        outer_idx = 0
        for reg in self.qubit_regs:
            if reg.name == qubit_id:
                return outer_idx
            outer_idx += reg.size
        raise LangException(f'Unable to find qubit register id: {qubit_id}.')

    def convert_qubit_id_to_indices(self, qubit_id: str) -> list[int]:
        outer_idx = 0
        for reg in self.qubit_regs:
            if reg.name == qubit_id:
                return [i + outer_idx for i in range(reg.size)]
            outer_idx += reg.size
        raise LangException(f'Unable to find qubit register id: {qubit_id}.')


eval_locals = {
    'pi': np.pi,
    'sin': np.sin,
    'cos': np.cos,
    'tan': np.tan,
    'ln': np.log,
    'exp': np.exp,
}


def eval_exp_recurse(tree: lark.Tree) -> Any:
    if isinstance(tree, lark.Token):
        return str(tree)
    code = ''
    for op in tree.children:
        if isinstance(op, lark.Token):
            code += str(op)
            continue
        if op.data == 'unaryexp':
            unaryop = op.children[0]
            code += f'{unaryop.children[0]}({eval_exp_recurse(op.children[1])})'
            continue
        elif op.data == 'usub':
            code += '-'
        elif op.data == 'pow':
            base = eval_exp_recurse(op.children[0])
            exp = eval_exp_recurse(op.children[1])
            code += f'{base}**{exp}'
            continue
        code += ' '.join(map(eval_exp_recurse, op.children))
    return code


def eval_exp(tree: lark.Tree) -> Any:
    return eval(eval_exp_recurse(tree), {}, eval_locals)


def eval_explist(tree: lark.Tree) -> Any:
    if isinstance(tree, list):
        tree = tree[0]
    if tree.data == 'explist':
        if len(tree.children) == 1:
            return eval_exp(tree.children[0])
        list_vals = eval_explist(tree.children[0])
        exp_vals = eval_exp(tree.children[1])
        if isinstance(list_vals, tuple):
            return (*list_vals, exp_vals)
        return list_vals, exp_vals
    elif tree.data == 'exp':
        return eval_exp(tree)
    else:
        raise Exception()


def qubits_from_list(tree: lark.Tree) -> list[tuple[str, int]]:
    if tree.data == 'anylist':
        return qubits_from_list(tree.children[0])
    elif tree.data == 'mixedlist':
        q2 = (str(tree.children[-2]), int(tree.children[-1]))
        if isinstance(tree.children[0], lark.Tree):
            q1 = qubits_from_list(tree.children[0])
            return [*q1, q2]
        return [q2]
    elif tree.data == 'argument':
        return [(str(tree.children[0]), int(tree.children[1]))]
    else:
        raise Exception()



"""This package contains integrations with the Qiskit framework."""
from __future__ import annotations

from bqskit.ext.qiskit.models import model_from_backend
from bqskit.ext.qiskit.translate import bqskit_to_qiskit
from bqskit.ext.qiskit.translate import qiskit_to_bqskit

__all__ = ['bqskit_to_qiskit', 'qiskit_to_bqskit', 'model_from_backend']



"""This package implements BQSKit's rule-based passes."""
from __future__ import annotations

from bqskit.passes.rules.ch2cnot import CHToCNOTPass
from bqskit.passes.rules.cnot2ch import CNOTToCHPass
from bqskit.passes.rules.cnot2cy import CNOTToCYPass
from bqskit.passes.rules.cnot2cz import CNOTToCZPass
from bqskit.passes.rules.cy2cnot import CYToCNOTPass
from bqskit.passes.rules.swap2cnot import SwapToCNOTPass
from bqskit.passes.rules.u3 import U3Decomposition
from bqskit.passes.rules.zxzxz import ZXZXZDecomposition

__all__ = [
    'CHToCNOTPass',
    'CNOTToCZPass',
    'CNOTToCHPass',
    'CNOTToCYPass',
    'CYToCNOTPass',
    'SwapToCNOTPass',
    'U3Decomposition',
    'ZXZXZDecomposition',
]



"""This module implements the QutritGate base class."""
from __future__ import annotations

from bqskit.ir.gate import Gate


class QutritGate(Gate):
    """A gate that only acts on qutrits."""

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return tuple([3] * self.num_qudits)



"""This module implements the CHGate."""
from __future__ import annotations

import math

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class CHGate(ConstantGate, QubitGate):
    """
    The controlled-Hadamard gate.

    The Controlled-H gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & \\frac{\\sqrt{2}}{2} & \\frac{\\sqrt{2}}{2} \\\\
        0 & 0 & \\frac{\\sqrt{2}}{2} & -\\frac{\\sqrt{2}}{2} \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'ch'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, math.sqrt(2) / 2, math.sqrt(2) / 2],
            [0, 0, math.sqrt(2) / 2, -math.sqrt(2) / 2],
        ],
    )



"""This module implements the SeedLayerGenerator class."""
from __future__ import annotations

import logging
from typing import cast
from typing import Sequence

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.point import CircuitPoint
from bqskit.passes.search.generator import LayerGenerator
from bqskit.passes.search.generators.simple import SimpleLayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence

_logger = logging.getLogger(__name__)


class SeedLayerGenerator(LayerGenerator):
    """Layer Generator for search that starts from a seed."""

    def __init__(
        self,
        seed: Circuit | Sequence[Circuit],
        forward_generator: LayerGenerator = SimpleLayerGenerator(),
        num_removed: int = 1,
        hash_on_1q_gate: bool = False,
    ) -> None:
        """
        Construct a SeedLayerGenerator.

        Args:
            seed (Circuit | Sequence[Circuit]): The seed or seeds
                to start synthesis from.

            forward_generator (LayerGenerator): A generator used to grow
                the circuit. (Default: SimpleLayerGenerator)

            num_removed (int): The number of atomic gate units removed
                from the circuit in each backwards branch. If 0, no
                backwards traversal of the synthesis tree will be done.
                (Default: 1)

            hash_on_1q_gate (bool): If True, 1 qubit gates that appear
                in circuit templates will be used in circuit hash
                calculations. Otherwise only multi-qubit gates will be
                used. Setting this value to False can help stability
                when using more complex LayerGenerators for the
                forward_generator. (Default: False)

        Raises:
            ValueError: If 'num_removed' is negative.
        """
        if not isinstance(forward_generator, LayerGenerator):
            raise TypeError(
                'Expected LayerGenerator for forward_generator'
                f', got {type(forward_generator)}.',
            )

        if not is_integer(num_removed):
            raise TypeError(
                'Expected integer for num_removed, '
                f'got {type(num_removed)}.',
            )
        if not isinstance(hash_on_1q_gate, bool):
            raise TypeError(
                'Expected bool for hash_on_1q_gate, '
                f'got {type(hash_on_1q_gate)}.',
            )

        if num_removed < 0:
            raise ValueError(
                'Expected non-negative value for num_removed, '
                f'got {num_removed}.',
            )

        if isinstance(seed, Circuit):
            seed = [seed]

        if not is_sequence(seed):
            raise TypeError(
                'Expected a Circuit or Sequence of Circuits for seed, '
                f'got {type(seed)}.',
            )

        if not all(isinstance(s, Circuit) for s in seed):
            msg = 'Expected a Circuit or Sequence of Circuits for seed.'
            for i, s in enumerate(seed):
                if not isinstance(s, Circuit):
                    msg += f'\nGot Sequence with {type(s)} at index {i}.'
            raise TypeError(msg)

        self.seeds = list(cast(Sequence[Circuit], seed))
        self.forward_generator = forward_generator
        self.num_removed = num_removed
        self.hash_on_1q_gate = hash_on_1q_gate

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            TypeError: If target is not a UnitaryMatrix, StateVector, or
                StateSystem.

        Notes:
            - For seeded layer generation, the initial_layer is the
              empty Circuit. The `gen_successors` method checks for this
              before returning `self.seeds` as successors.
        """
        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(f'Expected unitary or state, got {type(target)}.')

        return Circuit(target.num_qudits, target.radixes)

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        If `circuit` is the empty Circuit, seeds with radixes matching
        `circuit` will be used. If no seeds match, the empty Circuit will
        be returned as the only successor with a warning message.
        """
        if not isinstance(circuit, Circuit):
            raise TypeError(f'Expected Circuit, got {type(circuit)}.')

        if 'seed_seen_before' not in data:
            data['seed_seen_before'] = set()

        # If circuit is empty Circuit, successors are the seeds
        if circuit.is_empty:
            circ_hash = self.hash_structure(circuit, self.hash_on_1q_gate)
            # Do not return seeds if empty circuit already visited
            if circ_hash in data['seed_seen_before']:
                return []
            data['seed_seen_before'] = {circ_hash}

            # Check if any seeds match circuit, only use those seeds
            usable_seeds = [
                seed for seed in self.seeds
                if circuit.radixes == seed.radixes
            ]
            for seed in usable_seeds:
                h = self.hash_structure(seed, self.hash_on_1q_gate)
                data['seed_seen_before'].add(h)

            if len(usable_seeds) == 0:
                _logger.warning(
                    'No seeds matching the circuit\'s radixes found.'
                    '\nGenerating successors from empty circuit.'
                    '\nThis may cause a malformed search tree because the'
                    'generator never generated a proper initial layer.',
                )
                usable_seeds.append(circuit)

            return usable_seeds

        # Generate successors
        successors = self.forward_generator.gen_successors(circuit, data)

        # Search reverse direction
        ancestor_circuits = self.remove_atomic_units(circuit)
        successors = ancestor_circuits + successors

        # Filter out successors that have already been visited
        filtered_successors = []
        for s in successors:
            h = self.hash_structure(s, self.hash_on_1q_gate)
            if h not in data['seed_seen_before']:
                data['seed_seen_before'].add(h)
                filtered_successors.append(s)

        return filtered_successors

    @staticmethod
    def hash_structure(circuit: Circuit, hash_on_1q_gate: bool = True) -> int:
        """
        Compute a hash of the Circuit's structure.

        Args:
            circuit (Circuit): Circuit to be hashed.

            hash_on_1q_gate (bool): Whether or not to include 1 qubit
                gates in the hash computation. (Default: True)
        """
        hashes = []
        for cycle, op in circuit.operations_with_cycles():
            if not hash_on_1q_gate and op.num_qudits <= 1:
                continue
            hashes.append(hash((cycle, str(op))))
            if len(hashes) > 100:
                hashes = [sum(hashes)]
        return sum(hashes)

    def remove_atomic_units(self, circuit: Circuit) -> list[Circuit]:
        """
        Return circuits after removing upto `self.num_removed` atomic units.

        Atomic units defined as multi-qudit gates and the single-qudit
        gates that are directly dependent on them. Except for single-qudit
        circuits, where the single-qudit gate is the atomic unit.

        For two qudit synthesis, these atomic units look like:

            -- two_qudit_gate -- single_qudit_gate_1 --
                    |
            -- two_qudit_gate -- single_qudit_gate_2 --

        Generally, this will find the last `num_removed` multi-qudit
        gates, and remove them and any single qudit gates that are
        directly dependent on them.
        """
        num_removed = 0
        ancestor_circuits = []

        circuit_copy = circuit.copy()

        if circuit_copy.num_qudits == 1:
            init_num_cycles = circuit_copy.num_cycles
            for _ in range(min(self.num_removed, init_num_cycles)):
                circuit_copy.pop_cycle(-1)
                ancestor_circuits.append(circuit_copy.copy())

        for cycle, op in circuit.operations_with_cycles(reverse=True):

            if num_removed >= self.num_removed:
                break
            if op.num_qudits == 1:
                continue

            # Remove multi-qudit gate and single qudit dependents
            point = CircuitPoint(cycle, op.location[0])
            dependents = []
            for next_point in circuit_copy.next(point):
                if circuit_copy.get_operation(next_point).num_qudits == 1:
                    dependents.append(next_point)
            to_remove = dependents + [point]

            circuit_copy.batch_pop(to_remove)

            ancestor_circuits.append(circuit_copy.copy())
            num_removed += 1

        return ancestor_circuits



"""This module implements the ResidualFunction base classes."""
from __future__ import annotations

import abc

import numpy as np
import numpy.typing as npt

from bqskit.ir.opt.cost.function import CostFunction
from bqskit.qis.unitary.unitary import RealVector
from bqskit.utils.typing import is_real_number
from bqskit.utils.typing import is_sequence


class ResidualsFunction(CostFunction):
    """
    The ResidualFunction base class.

    A ResidualFunction is a map from a vector of real numbers to a vector of
    real number. This output represents the residuals and minimizers will
    attempt to reduce it. The input should represent circuit parameters.
    """

    @abc.abstractmethod
    def get_residuals(self, params: RealVector) -> npt.NDArray[np.float64]:
        """Return the vector of residuals given the input parameters."""

    def __call__(  # type: ignore
        self,
        params: RealVector,
    ) -> npt.NDArray[np.float64]:
        """Return the vector of residuals given the input parameters."""

        if not is_sequence(params):
            raise TypeError(
                'Expected sequence for params, got %s.' % type(params),
            )

        if not all(is_real_number(param) for param in params):
            raise TypeError(
                'Expected sequence of floats for params.',
            )

        return self.get_residuals(params)



"""This module implements the SetModelPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.machine import MachineModel
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.qis.graph import CouplingGraph

_logger = logging.getLogger(__name__)


class SetModelPass(BasePass):
    """Sets a target machine model for future passes to target."""

    def __init__(self, model: MachineModel) -> None:
        """
        Construct a SetModelPass.

        Args:
            model (MachineModel | None): The machine model to encode
                into the compilation workflow.
        """
        if not isinstance(model, MachineModel):
            raise TypeError(f'Expected MachineModel, got: {type(model)}.')

        self.model = model

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if self.model.num_qudits < circuit.num_qudits:
            raise RuntimeError('Machine model is too small for circuit.')

        data.model = self.model  # Update Model
        data.placement = list(range(circuit.num_qudits))  # Reset placement


class ExtractModelConnectivityPass(BasePass):
    """
    Extracts and saves the current target machine model's connectivity.

    The model will remain unchanged except that it will be fully connected until
    the RestoreModelConnectivityPass is executed.
    """

    key = '_ExtractModelConnectivityPass_connectivity'

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        data[self.key] = data.model.coupling_graph
        data.model.coupling_graph = CouplingGraph.all_to_all(
            data.model.num_qudits,
        )


class RestoreModelConnectivityPass(BasePass):
    """Restores the connectivity of the target machine model."""

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        if ExtractModelConnectivityPass.key not in data:
            raise RuntimeError(
                'Cannot restore connectivity without first '
                'extracting it using ExtractModelConnectivityPass.',
            )

        data.model.coupling_graph = data[ExtractModelConnectivityPass.key]
        del data[ExtractModelConnectivityPass.key]



"""This module implements the PermutationAwareSynthesisPass pass."""
from __future__ import annotations

import itertools as it
import logging
from typing import Callable

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.passes.synthesis.leap import LEAPSynthesisPass
from bqskit.passes.synthesis.synthesis import SynthesisPass
from bqskit.qis.permutation import PermutationMatrix
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.runtime import get_runtime


_logger = logging.getLogger(__name__)


def op_count(circuit: Circuit) -> float:
    """Counts the number of operations in a circuit."""
    return float(circuit.num_operations)


class PermutationAwareSynthesisPass(SynthesisPass):
    """Synthesis pass considering target permutations."""

    def __init__(
        self,
        input_perm: bool = False,
        output_perm: bool = True,
        inner_synthesis: SynthesisPass = LEAPSynthesisPass(),
        scoring_fn: Callable[[Circuit], float] = op_count,
    ) -> None:
        """
        Construct a PermutationAwareSynthesisPass.

        Args:
            input_perm (bool): If true, vary the input permutation
                during synthesis. (Default: False)

            output_perm (bool): If true, vary the output permutation
                during synthesis. (Default: True)

            inner_synthesis (SynthesisPass): The synthesis algorithm used
                on all permutations. (Default: :class:`LEAPSynthesisPass`)

            scoring_fn (Callable[[Circuit], float]): The function used to
                score the permuted circuits. The smallest score wins.
                (Default: :func:`op_count`)
        """
        # TODO: Add option to choose first returned result
        if not isinstance(inner_synthesis, SynthesisPass):
            bad_type = type(inner_synthesis)
            raise TypeError(f'Expected SynthesisPass object, got {bad_type}.')

        if not callable(scoring_fn):
            bad_type = type(scoring_fn)
            m = f'Expected a function from circuits to scores, got {bad_type}.'
            raise TypeError(m)

        self.inner_synthesis = inner_synthesis
        self.scoring_fn = scoring_fn
        self.input_perm = input_perm
        self.output_perm = output_perm

    async def synthesize(
        self,
        utry: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """Synthesize `utry`, see :class:`SynthesisPass` for more."""
        if not all(r == utry.radixes[0] for r in utry.radixes):
            raise NotImplementedError(
                'PermutationAwareSynthesisPass only supports unitaries '
                'with the same radix on all qudits currently.',
            )

        # Calculate all permuted targets
        width = utry.num_qudits
        perms = list(it.permutations(range(width)))
        no_perm = [tuple(range(width))]
        Pis = [
            PermutationMatrix.from_qudit_location(width, utry.radixes[0], p)
            for p in perms
        ]
        Pos = [
            PermutationMatrix.from_qudit_location(width, utry.radixes[0], p)
            for p in perms
        ]

        if self.input_perm and self.output_perm:
            permsbyperms = list(it.product(perms, perms))
            targets = [Po.T @ utry @ Pi for Pi, Po in it.product(Pis, Pos)]

        elif self.input_perm:
            permsbyperms = list(it.product(perms, no_perm))
            targets = [utry @ Pi for Pi in Pis]

        elif self.output_perm:
            permsbyperms = list(it.product(no_perm, perms))
            targets = [Po.T @ utry for Po in Pos]

        else:
            _logger.warning('No permutation is being used in PAS.')
            permsbyperms = list(it.product(no_perm, no_perm))
            targets = [utry]

        # Synthesize all permuted targets
        circuits: list[Circuit] = await get_runtime().map(
            self.inner_synthesis.synthesize,
            targets,
            [data] * len(targets),
            log_context=[{'perm': str(perm)} for perm in permsbyperms],
        )

        # Return best circuit
        best_circuit = circuits[0]
        best_perm = permsbyperms[0]
        best_score = self.scoring_fn(circuits[0])
        for perm, circuit in zip(permsbyperms[1:], circuits[1:]):
            score = self.scoring_fn(circuit)
            if score < best_score:
                best_circuit = circuit
                best_perm = perm
                best_score = score

        data['initial_mapping'] = best_perm[0]
        data['final_mapping'] = best_perm[1]
        return best_circuit



"""This package contains composed gates."""
from __future__ import annotations

from bqskit.ir.gates.composed.controlled import ControlledGate
from bqskit.ir.gates.composed.daggergate import DaggerGate
from bqskit.ir.gates.composed.embedded import EmbeddedGate
from bqskit.ir.gates.composed.frozenparam import FrozenParameterGate
from bqskit.ir.gates.composed.powergate import PowerGate
from bqskit.ir.gates.composed.tagged import TaggedGate
from bqskit.ir.gates.composed.vlg import VariableLocationGate

__all__ = [
    'ControlledGate',
    'PowerGate',
    'DaggerGate',
    'EmbeddedGate',
    'FrozenParameterGate',
    'TaggedGate',
    'VariableLocationGate',
]



"""This module implements the PAMRoutingPass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.point import CircuitPoint
from bqskit.passes.control.foreach import ForEachBlockPass
from bqskit.passes.mapping.pam import PAMBlockTAPermData
from bqskit.passes.mapping.pam import PermutationAwareMappingAlgorithm

_logger = logging.getLogger(__name__)


class PAMRoutingPass(PermutationAwareMappingAlgorithm, BasePass):

    out_data_key = '_pam_routing_block_out_data'

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        subgraph = data.connectivity
        if not subgraph.is_fully_connected():
            raise RuntimeError('Cannot route circuit on disconnected qudits.')

        perm_data: dict[CircuitPoint, PAMBlockTAPermData] = {}
        block_datas = data[ForEachBlockPass.key][-1]
        for block_data in block_datas:
            perm_data[block_data['point']] = block_data['permutation_data']

        pi = [i for i in range(circuit.num_qudits)]
        out_data = self.forward_pass(circuit, pi, subgraph, perm_data, True)
        data.final_mapping = [pi[x] for x in data.final_mapping]

        _logger.info(f'Finished routing with layout: {str(pi)}')
        data[self.out_data_key] = out_data



"""
This test module verifies all conftest fixtures work as intended.

The BQSKit root conftest defines the following circuit fixtures that
need to be verified:
    gen_random_utry_np
    gen_invalid_utry_np
    gen_random_circuit

There are four sets of randomly-generated circuits that are used to test
`gen_random_circuit` by proxy:
    r3_qubit_circuit
    r3_qubit_constant_circuit
    r3_qutrit_circuit
    r6_qudit_circuit
"""
from __future__ import annotations

from typing import Any

import numpy as np

from bqskit.ir.circuit import Circuit
from bqskit.qis.unitary import UnitaryMatrix


class TestGenRandomUtryNp:
    """Ensure random utry generator behaves as expected."""

    def test_invalid_1(self, gen_random_utry_np: Any) -> None:
        try:
            utry = gen_random_utry_np('a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_2(self, gen_random_utry_np: Any) -> None:
        try:
            utry = gen_random_utry_np(['a', 3])  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_valid_single_dim(self, gen_random_utry_np: Any) -> None:
        utry = gen_random_utry_np(8)
        assert isinstance(utry, np.ndarray)
        assert utry.shape == (8, 8)
        assert UnitaryMatrix.is_unitary(utry)

    def test_valid_multi_dim(self, gen_random_utry_np: Any) -> None:
        utry = gen_random_utry_np([4, 8])
        assert isinstance(utry, np.ndarray)
        assert utry.shape == (8, 8) or utry.shape == (4, 4)
        assert UnitaryMatrix.is_unitary(utry)


class TestGenInvalidUtryNp:
    """Ensure invalid utry generator behaves as expected."""

    def test_invalid_1(self, gen_invalid_utry_np: Any) -> None:
        try:
            iutry = gen_invalid_utry_np('a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_2(self, gen_invalid_utry_np: Any) -> None:
        try:
            iutry = gen_invalid_utry_np(['a', 3])  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_valid_single_dim(self, gen_invalid_utry_np: Any) -> None:
        iutry = gen_invalid_utry_np(8)
        assert isinstance(iutry, np.ndarray)
        assert iutry.shape == (8, 8)
        assert not UnitaryMatrix.is_unitary(iutry)

    def test_valid_multi_dim(self, gen_invalid_utry_np: Any) -> None:
        iutry = gen_invalid_utry_np([4, 8])
        assert isinstance(iutry, np.ndarray)
        assert iutry.shape == (8, 8) or iutry.shape == (4, 4)
        assert not UnitaryMatrix.is_unitary(iutry)


class TestGenRandomCircuit:
    """Ensure random circuit generator behaves as expected."""

    def test_invalid_type_1(self, gen_random_circuit: Any) -> None:
        try:
            circuit = gen_random_circuit('a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_type_2(self, gen_random_circuit: Any) -> None:
        try:
            circuit = gen_random_circuit(3, 'a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_type_3(self, gen_random_circuit: Any) -> None:
        try:
            circuit = gen_random_circuit(3, [2, 2, 2], 'a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_type_4(self, gen_random_circuit: Any) -> None:
        try:
            circuit = gen_random_circuit(3, [2, 2, 2], 5, 'a')  # noqa
        except TypeError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    def test_invalid_value_1(self, gen_random_circuit: Any) -> None:
        try:
            circuit = gen_random_circuit(3, [2, 2])  # noqa
        except ValueError:
            return
        except BaseException:
            assert False, 'Unexpected error.'

    # r3_qubit_circuit
    def test_r3_size(self, r3_qubit_circuit: Circuit) -> None:
        assert r3_qubit_circuit.num_qudits == 3

    def test_r3_radix(self, r3_qubit_circuit: Circuit) -> None:
        assert r3_qubit_circuit.is_qubit_only()

    def test_r3_depth(self, r3_qubit_circuit: Circuit) -> None:
        assert r3_qubit_circuit.num_operations == 10

    # r3_qubit_constant_circuit
    def test_r3_con_size(self, r3_qubit_constant_circuit: Circuit) -> None:
        assert r3_qubit_constant_circuit.num_qudits == 3

    def test_r3_con_radix(self, r3_qubit_constant_circuit: Circuit) -> None:
        assert r3_qubit_constant_circuit.is_qubit_only()

    def test_r3_con_depth(self, r3_qubit_constant_circuit: Circuit) -> None:
        assert r3_qubit_constant_circuit.num_operations == 25

    def test_r3_con_constant(self, r3_qubit_constant_circuit: Circuit) -> None:
        assert r3_qubit_constant_circuit.is_constant()

    # r3_qutrit_circuit
    def test_r3_qutrit_size(self, r3_qutrit_circuit: Circuit) -> None:
        assert r3_qutrit_circuit.num_qudits == 3

    def test_r3_qutrit_radix(self, r3_qutrit_circuit: Circuit) -> None:
        assert r3_qutrit_circuit.is_qutrit_only()

    def test_r3_qutrit_depth(self, r3_qutrit_circuit: Circuit) -> None:
        assert r3_qutrit_circuit.num_operations == 10

    # r6_qudit_circuit
    def test_r6_size(self, r6_qudit_circuit: Circuit) -> None:
        assert r6_qudit_circuit.num_qudits == 6

    def test_r6_radix(self, r6_qudit_circuit: Circuit) -> None:
        count = r6_qudit_circuit.radixes.count(2)
        count += r6_qudit_circuit.radixes.count(3)
        assert count == r6_qudit_circuit.num_qudits

    def test_r6_depth(self, r6_qudit_circuit: Circuit) -> None:
        assert r6_qudit_circuit.num_operations == 10



"""This module implements the VariableLocationGate."""
from __future__ import annotations

from typing import cast
from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.qis.permutation import PermutationMatrix
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.math import softmax


class VariableLocationGate(ComposedGate):
    """
    Gate that can multiplex multiple placements of another gate.

    A VariableLocationGate continuously interpolates between multiple locations
    for another gate. To do this this composed gate becomes as large as the sum
    of all locations.
    """

    def __init__(
        self,
        gate: Gate,
        locations: Sequence[CircuitLocationLike],
        radixes: Sequence[int] = [],
    ) -> None:
        """
        Create a gate that has parameterized location.

        Args:
            gate (Gate): The gate to parameterize location for.

            locations (Sequence[CircuitLocationLike]): A sequence of locations.
                Each location represents a valid placement for gate.

            radixes (Sequence[int]): The number of orthogonal
                states for each qudit. Attempts to infer it from `gate`
                and `locations` if not specified.

        Raises:
            ValueError: If there are not enough locations or the locations
                are incorrectly sized.

        Notes:
            The locations are calculated in their own space and are not
            relative to a circuit. This means you should consider the
            VariableLocationGate as its own circuit when deciding the
            locations. For example, if you want to multiplex the (2, 3)
            and (3, 5) placements of a CNOT on a 6-qubit circuit, then
            you would give the VariableLocationGate the (0, 1) and (1, 2)
            locations and place the VariableLocationGate on qubits
            (2, 3, 5) on the circuit.
        """
        if not isinstance(gate, Gate):
            raise TypeError('Expected gate object, got %s' % type(gate))

        if not all(CircuitLocation.is_location(l) for l in locations):
            raise TypeError('Expected a sequence of valid locations.')

        locations = [CircuitLocation(l) for l in locations]

        if not all(len(l) == gate.num_qudits for l in locations):
            raise ValueError('Invalid sized location.')

        if len(locations) < 1:
            raise ValueError('VLGs require at least 1 locations.')

        self.gate = gate
        name_str_tuple = (gate.name, locations, radixes)
        self._name = 'VariableLocationGate(%s, %s, %s)' % name_str_tuple
        self.locations = list(locations)
        self._num_qudits = len(set(sum((tuple(l) for l in locations), tuple())))

        if len(radixes) == 0:
            # Calculate radixes
            radix_map: dict[int, int | None] = {
                i: None for i in range(self.num_qudits)
            }
            for l in locations:
                for radix, qudit_index in zip(gate.radixes, l):
                    if radix_map[qudit_index] is None:
                        radix_map[qudit_index] = radix
                    elif radix_map[qudit_index] != radix:
                        raise ValueError(
                            'Gate cannot be applied to all locations'
                            ' due to radix mismatch.',
                        )

            if None in radix_map.values():
                raise ValueError('VariableUnitaryGate cannot infer radixes.')

            self._radixes = tuple(radix_map.values())  # type: ignore
        else:
            for l in locations:
                for radix, qudit_index in zip(gate.radixes, l):
                    if radixes[qudit_index] != radix:
                        raise ValueError(
                            'Gate cannot be applied to all locations'
                            ' due to radix mismatch.',
                        )

            self._radixes = tuple(radixes)
            self._num_qudits = len(self.radixes)

        self._num_params = self.gate.num_params + len(locations)

        self.extension_size = self.num_qudits - self.gate.num_qudits
        # TODO: This needs to changed for radixes
        self.I = np.identity(2 ** self.extension_size)
        self.perms = np.array([
            PermutationMatrix.from_qubit_location(self.num_qudits, l)
            for l in self.locations
        ])

    def get_location(self, params: RealVector) -> tuple[int, ...]:
        """Returns the gate's location."""
        idx = int(np.argmax(self.split_params(params)[1]))
        return tuple(self.locations[idx])

    def split_params(
        self,
        params: RealVector,
    ) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
        """Split params into subgate params and location params."""
        return (
            np.array(params[:self.gate.num_params]),
            np.array(params[self.gate.num_params:]),
        )

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        self.check_parameters(params)
        a, l = self.split_params(params)
        l = softmax(l, 10)

        P = np.sum([a * s for a, s in zip(l, self.perms)], 0)
        G = self.gate.get_unitary(a)
        PGPT = P.T @ np.kron(G, self.I) @ P
        return UnitaryMatrix.closest_to(PGPT, self.radixes)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        return self.get_unitary_and_grad(params)[1]

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        self.check_parameters(params)
        a, l = self.split_params(params)
        l = softmax(l, 10)

        P = np.sum([a * s for a, s in zip(l, self.perms)], 0)
        G = self.gate.get_unitary(a)
        G = np.kron(G, self.I)
        PG = P @ G
        GPT = G @ P.T
        PGPT = P @ GPT

        dG = cast(DifferentiableUnitary, self.gate).get_grad(a)
        dG = np.kron(dG, self.I)
        dG = P @ dG @ P.T

        perm_array = np.array([perm for perm in self.perms])
        dP = perm_array @ GPT + PG @ perm_array.transpose((0, 2, 1)) - 2 * PGPT
        dP = np.array([10 * x * y for x, y in zip(l, dP)])
        U = UnitaryMatrix.closest_to(PGPT, self.radixes)
        return U, np.concatenate([dG, dP])

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, VariableLocationGate)
            and self.gate == other.gate
            and self.locations == other.locations
            and self.radixes == other.radixes
        )

    def __hash__(self) -> int:
        return hash((self.gate, tuple(self.locations), self.radixes))



"""This module implements the DiagonalPredicate class."""
from __future__ import annotations

from typing import TYPE_CHECKING

from bqskit.passes.control.predicate import PassPredicate
from bqskit.utils.math import diagonal_distance

if TYPE_CHECKING:
    from bqskit.compiler.passdata import PassData
    from bqskit.ir.circuit import Circuit


class DiagonalPredicate(PassPredicate):
    """
    The DiagonalPredicate class.

    The DiagonalPredicate class returns True if the circuit's unitary can be
    approximately inverted by a diagonal unitary. A unitary is approximately
    inverted when the Hilbert-Schmidt distance to the identity is less than some
    threshold.
    """

    def __init__(self, threshold: float) -> None:
        """
        Construct a DiagonalPredicate.

        Args:
            threshold (float): If a circuit can be approximately inverted
                by a diagonal unitary (meaning the Hilbert-Schmidt distance
                to the identity is less than or equal to this number after
                multiplying by the diagonal unitary), True is returned.
        """
        self.threshold = threshold

    def get_truth_value(self, circuit: Circuit, data: PassData) -> bool:
        """Call this predicate, see :class:`PassPredicate` for more info."""
        dist = diagonal_distance(circuit.get_unitary().numpy)
        return dist <= self.threshold



"""This module implements the SingleQuditLayerGenerator class."""
from __future__ import annotations

import logging
from typing import Sequence

from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.passes.search.generator import LayerGenerator
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.system import StateSystem
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.typing import is_sequence
_logger = logging.getLogger(__name__)


class SingleQuditLayerGenerator(LayerGenerator):
    """
    The SingleQuditLayerGenerator class.

    This module implements a layer generator for a single-qudit circuits with
    hetergenous gate support.
    """

    def __init__(
        self,
        gates: Sequence[Gate] | None,
        allow_repeats: bool = False,
    ) -> None:
        """
        Construct a SingleQuditLayerGenerator.

        Args:
            gates (Sequence[Gate] | None): The single-qudit gate list. If None,
                then the generator will read the pass data for the model's
                gate set.

            allow_repeats (bool): By default this is set to False,
                and the generator will not generate a circuit that
                has the same gate twice in a row. Some gates are not
                closed under multiplication and it may be beneficial
                to allow them to repeat. In this case, `allow_repeats`
                should be True.

        Raises:
            ValueError: If any pair of gates from `gates` has a
                different radix.

            ValueError: If any gate from `gates` is not single-qudit.

            ValueError: If `gates` is empty.
        """
        if gates is not None:
            self.radix = None
            if not is_sequence(gates):
                raise TypeError(
                    f'Expected sequence of gates, got {type(gates)}.',
                )

            for gate in gates:
                if not isinstance(gate, Gate):
                    raise TypeError(f'Expected gate, got {type(gate)}.')

                if gate.num_qudits > 1:
                    raise ValueError(f'Expected single-qudit gate, got {gate}.')

                if self.radix is None:
                    self.radix = gate.radixes[0]

                elif gate.radixes[0] != self.radix:
                    raise ValueError('Gate set has a radix inconsistency.')

            if len(gates) == 0:
                raise ValueError('Empty gate set.')

        self.gates = gates
        self.allow_repeats = allow_repeats

    def gen_initial_layer(
        self,
        target: UnitaryMatrix | StateVector | StateSystem,
        data: PassData,
    ) -> Circuit:
        """
        Generate the initial layer, see LayerGenerator for more.

        Raises:
            ValueError: If `target` has a radix mismatch with
                `self.gates`.

            ValueError: If `target` is not single-qudit.
        """
        if self.gates is None:
            self.gates = list(data.gate_set.single_qudit_gates)
            self.radix = self.gates[0].radixes[0]
            if not all(g.radixes[0] == self.radix for g in self.gates):
                raise RuntimeError('Gate set has a radix inconsistency.')

        if not isinstance(target, (UnitaryMatrix, StateVector, StateSystem)):
            raise TypeError(
                'Expected unitary or state, got %s.' % type(target),
            )

        if target.num_qudits > 1:
            raise ValueError('Target is larger than a single-qudit.')

        if target.radixes[0] != self.radix:
            raise ValueError('Mismatch between generator and target radix.')

        init_circuit = Circuit(target.num_qudits, target.radixes)
        return init_circuit

    def gen_successors(self, circuit: Circuit, data: PassData) -> list[Circuit]:
        """
        Generate the successors of a circuit node.

        Raises:
            ValueError: If circuit is not a single-qudit circuit.
        """

        if not isinstance(circuit, Circuit):
            raise TypeError('Expected circuit, got %s.' % type(circuit))

        if circuit.num_qudits > 1:
            raise ValueError('Cannot expand a multi-qudit circuit.')

        # Record last gate
        last_gate = circuit[-1, -1].gate if len(circuit) >= 1 else None

        # Generate successors
        successors = []
        for gate in self.gates:  # type: ignore  # self.gates is set earlier

            # Skip adding a gate that repeats the last one on the circuit
            if not self.allow_repeats and gate == last_gate:
                continue

            successor = circuit.copy()
            successor.append_gate(gate, 0)
            successors.append(successor)

        return successors



"""The state package exports classes related to state vectors."""
from __future__ import annotations

from bqskit.qis.state.state import StateLike
from bqskit.qis.state.state import StateVector
from bqskit.qis.state.statemap import StateVectorMap
from bqskit.qis.state.system import StateSystem
from bqskit.qis.state.system import StateSystemLike

__all__ = [
    'StateVector',
    'StateLike',
    'StateVectorMap',
    'StateSystem',
    'StateSystemLike',
]



"""This module implements the ControlledGate class."""
from __future__ import annotations

from typing import cast
from typing import Sequence

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs
from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_sequence
from bqskit.utils.typing import is_sequence_of_int
from bqskit.utils.typing import is_valid_radixes


class EmbeddedGate(ComposedGate, DifferentiableUnitary):
    """
    An embedding of a gate into a higher-dimensional qudit gate.

    For example, a qubit gate can be embedded into a qutrit gate by
    mapping all qubit levels to a subspace of the qutrit levels.

    This transformation can be shown directly on unitaries. If we have an
    arbitrary single-qubit gate :math:`U` given by the following matrix:

    .. math::

        U = \\begin{pmatrix}
            a & b \\\\
            c & d \\\\
        \\end{pmatrix}

    and we want to embed this into the 0 and 2 levels of a qutrit gate,
    then we can do so by mapping the qubit levels 0 and 1 to the qutrit
    levels 0 and 2, respectively. This gives us the following matrix:

    .. math::

        U_{embedded} = \\begin{pmatrix}
            a & 0 & b \\\\
            0 & 1 & 0 \\\\
            c & 0 & d \\\\
        \\end{pmatrix}

    This concept can be generalized to multiple qudits and even
    mixed-radix systems.

    Note:
        - Global phase inconsistencies in gates will become local phase
            inconsistencies in the embedded gate. For example, if the
            global phase difference between the U1Gate and the RZGate
            will become local phase differences in the corresponding
            subspaces when embedded into a higher-dimensional qudit.
    """

    def __init__(
        self,
        gate: Gate,
        radixes: Sequence[int] | int,
        level_maps: None | Sequence[int] | Sequence[Sequence[int]] = None,
    ) -> None:
        """
        Construct an EmbeddedGate.

        Args:
            gate (Gate): The gate to embed in a higher-dimensional qudit
                gate.

            radixes (Sequence[int] | int): The target radixes of the higher-
                dimensional system. If an integer is given, then the radixes
                are assumed to be the same for all qudits. For example,
                if `radixes = 3`, then the gate will be embedded into a
                qutrit gate.

             level_maps (None | Sequence[int] | Sequence[Sequence[int]]):
                 The level map for the embedding for each qudit. If a
                 sequence of integers is given, then the level map is
                 assumed to be the same for all qudits. For example, if
                 `radixes = 3` and `level_maps = [0, 2]`, then the gate
                 will be embedded into a qutrit gate by mapping the qubit
                 levels 0 and 1 to the qutrit levels 0 and 2,
                 respectively. If a sequence of sequences is given, then
                 the level map is assumed to be different for each qudit.
                 For example, if `radixes = [3, 3]` and `level_maps =
                 [[0, 2], [1, 2]]`, then the gate will be embedded into a
                 two-qudit gate by mapping the first qubit's 0 and 1
                 levels to the first qutrit's 0 and 2 levels,
                 respectively, and by mapping the second qubit's 0 and 1
                 levels to the second qutrit's 1 and 2 levels,
                 respectively. This can also be set to `None`, which will
                 embed the lower dimension gate in the lowest levels of
                 the new radixes.

        Raises:

            ValueError: If any radix is less than 2.

            ValueError: If radixes is given as a sequence and its length
                is not equal to the number of qudits in the gate.

            ValueError: If any of the gate's radixes are greater than the
                corresponding target radixes.

            ValueError: If the level map is given as a sequence of sequences
                and its length is not equal to the number of qudits in the
                gate.

            ValueError: If any of the individual qudit level maps are not
                the same length as the gate's corresponding qudit radix.

            ValueError: If any individual qudit level map has an invalid
                qudit level, i.e. too low (< 0) or too high (>= radix).

            ValueError: If any individual qudit level map is not one-to-one,
                i.e. if any two qudit levels are mapped to the same target
                qudit level.

        Examples: (#TODO update)
            XGate for qutrits:
                ```
                > x_qutrit_gate = ControlledGate(XGate(),[3],[0,2])
                > x_qutrit_gate.get_unitary()
                > [[0, 0, sqrt(2)/2],
                   [0, 0, 0],
                   [sqrt(2)/2, 0, 0]]
                ```
        """
        if not isinstance(gate, Gate):
            raise TypeError(f'Expected gate object, got {type(gate)}.')

        if is_integer(radixes):
            radixes = [radixes] * gate.num_qudits

        radixes = cast(Sequence[int], radixes)

        if not is_valid_radixes(radixes, gate.num_qudits):
            raise ValueError(
                'Given target radixes was not valid. Either invalid type,'
                ' invalid length, or invalid radix. Expected target radixes'
                ' to be a single integer or sequence of integers with length'
                f' equal to {gate.num_qudits=}. Also expected every radix'
                f' to be greater than 2, got {radixes=}.',
            )

        if level_maps is None:
            level_maps = [list(range(levels)) for levels in gate.radixes]

        if not is_sequence(level_maps):
            raise TypeError(
                'Expected level_maps to be a sequence of integers or a '
                f'sequence of sequences of integers, got {level_maps}.',
            )

        if is_sequence_of_int(level_maps):
            level_maps = [level_maps] * gate.num_qudits

        if not all(is_sequence_of_int(level_map) for level_map in level_maps):
            raise TypeError(
                'Expected level_maps to be a sequence of integers or a '
                f'sequence of sequences of integers, got {level_maps}.',
            )

        level_maps = cast(Sequence[Sequence[int]], level_maps)

        if any(gr > tr for gr, tr in zip(gate.radixes, radixes)):
            raise ValueError(
                'Given target radixes was not valid. Expected every target'
                ' radix to be greater than or equal to the corresponding'
                f' gate radix, got {gate.radixes=} and {radixes=}.',
            )

        if len(level_maps) != gate.num_qudits:
            raise ValueError(
                'Given level_maps was not valid. Expected level_maps to be'
                ' a sequence of sequences of ints with length equal to '
                f'{gate.num_qudits=}, got {len(level_maps)=}.',
            )

        if any(len(lmap) != gr for gr, lmap in zip(gate.radixes, level_maps)):
            raise ValueError(
                'Given level_maps was not valid. Expected every level_map'
                ' to have length equal to the corresponding gate radix, got'
                f' {gate.radixes=} and {level_maps=}.',
            )

        if any(
            any(lvl < 0 or lvl >= r for lvl in lmap)
            for r, lmap in zip(radixes, level_maps)
        ):
            raise ValueError(
                'Given level_maps was not valid. Expected every level_map'
                ' to have all levels in the range [0, radix), got'
                f' {radixes=} and {level_maps=}.',
            )

        if any(len(lmap) != len(set(lmap)) for lmap in level_maps):
            raise ValueError(
                'Given level_maps was not valid. Expected every level_map'
                f' to be one-to-one, got duplicate levels: {level_maps=}.',
            )

        self.gate = gate
        self.level_maps = tuple([tuple(list(lmap)) for lmap in level_maps])
        self._num_qudits = gate._num_qudits
        self._name = f'Embedded({self.gate.name}){self.level_maps}'
        self._num_params = self.gate._num_params
        self._radixes = tuple(radixes)
        self._dim = int(np.prod(self.radixes))

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0 and not building_docs():
            U = self.gate.get_unitary()
            U_embed = np.eye(self.dim, dtype=np.complex128)
            self._map_matrix(U, U_embed)
            self._utry = UnitaryMatrix(U_embed, self.radixes, False)

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, '_utry'):
            return self._utry

        U = self.gate.get_unitary(params)
        U_embed = np.eye(self.dim, dtype=np.complex128)
        self._map_matrix(U, U_embed)
        return UnitaryMatrix(U_embed, self.radixes, False)

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return np.array([])

        G = self.gate.get_grad(params)  # type: ignore
        G_embed = []
        for g in G:
            M = np.zeros((self.dim, self.dim), dtype=np.complex128)
            self._map_matrix(g, M)
            G_embed.append(M)
        return np.array(G_embed, dtype=np.complex128)

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, '_utry'):
            return self._utry, np.array([])

        U, G = self.gate.get_unitary_and_grad(params)  # type: ignore
        U_embed = np.eye(self.dim, dtype=np.complex128)
        self._map_matrix(U, U_embed)

        G_embed = []
        for g in G:
            M = np.zeros((self.dim, self.dim), dtype=np.complex128)
            self._map_matrix(g, M)
            G_embed.append(M)

        return (
            UnitaryMatrix(U_embed, self.radixes, False),
            np.array(G_embed, dtype=np.complex128),
        )

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, EmbeddedGate)
            and self.gate == other.gate
            and self.radixes == other.radixes
            and self.level_maps == other.level_maps
        )

    def __hash__(self) -> int:
        return hash((self.gate, self.radixes, self.level_maps))

    def _map_matrix(
        self,
        small: npt.NDArray[np.complex128] | UnitaryMatrix,
        big: npt.NDArray[np.complex128],
    ) -> None:
        """
        Map the lower-dimensional `small` matrix into the larger `big` matrix.

        This will mutate the `big` matrix in-place. The `small` matrix
        must be a square matrix with dimensions equal to the dimension
        of the gate being embedded. The `big` matrix must be a square
        matrix with dimensions equal to the dimension of the gate being
        embedded into.

        The embedding is done by mapping the indices of the `small`
        matrix to the indices of the `big` matrix using the level maps.

        When doing unitary calculations, pass the identity in for the
        `big` matrix and the unitary of the gate being embedded in for
        the `small` matrix.

        When doing gradient calculations, pass the zero matrix in for
        the `big` matrix and the gradient of the gate being embedded in
        for the `small` matrix.

        Args:
            small (npt.NDArray[np.complex128]): The matrix to embed.

            big (npt.NDArray[np.complex128]): The matrix to embed into.

        Notes:
            No checks are done to ensure that the parameters are correct.
        """

        for i in range(self.gate.dim):
            # Expand i, j into the mixed-radix basis of the gate
            i_exp = np.unravel_index(i, self.gate.radixes)

            # Map the indices to the target basis using the level maps
            i_map_exp = [lm[ie] for lm, ie in zip(self.level_maps, i_exp)]

            # Convert the mapped expanded indices back to a flat index
            i_target = np.ravel_multi_index(i_map_exp, self.radixes)

            for j in range(self.gate.dim):
                # Expand i, j into the mixed-radix basis of the gate
                j_exp = np.unravel_index(j, self.gate.radixes)

                # Map the indices to the target basis using the level maps
                j_map_exp = [lm[je] for lm, je in zip(self.level_maps, j_exp)]

                # Convert the mapped expanded indices back to a flat index
                j_target = np.ravel_multi_index(j_map_exp, self.radixes)
                big[i_target, j_target] = small[i, j]



"""This module implements the DaggerGate Class."""
from __future__ import annotations

import numpy as np
import numpy.typing as npt

from bqskit.ir.gate import Gate
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.qis.unitary.differentiable import DifferentiableUnitary
from bqskit.qis.unitary.optimizable import LocallyOptimizableUnitary
from bqskit.qis.unitary.unitary import RealVector
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
from bqskit.utils.docs import building_docs


class DaggerGate(
    ComposedGate,
    LocallyOptimizableUnitary,
    DifferentiableUnitary,
):
    """
    An arbitrary inverted gate.

    The DaggerGate is a composed gate that equivalent to the
    conjugate transpose of the input gate.

    For example:
        >>> from bqskit.ir.gates import TGate, TdgGate
        >>> DaggerGate(TGate()).get_unitary() == TdgGate().get_unitary()
        True
    """

    def __init__(self, gate: Gate) -> None:
        """
        Create a gate which is the conjugate transpose of another.

        Args:
            gate (Gate): The Gate to conjugate transpose.
        """
        if not isinstance(gate, Gate):
            raise TypeError('Expected gate object, got %s' % type(gate))

        self.gate = gate
        self._name = 'Dagger(%s)' % gate.name
        self._num_params = gate.num_params
        self._num_qudits = gate.num_qudits
        self._radixes = gate.radixes

        # If input is a constant gate, we can cache the unitary.
        if self.num_params == 0 and not building_docs():
            self.utry = gate.get_unitary().dagger

    def get_unitary(self, params: RealVector = []) -> UnitaryMatrix:
        """Return the unitary for this gate, see :class:`Unitary` for more."""
        if hasattr(self, 'utry'):
            return self.utry

        return self.gate.get_unitary(params).dagger

    def get_grad(self, params: RealVector = []) -> npt.NDArray[np.complex128]:
        """
        Return the gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.

        Notes:
            The derivative of the conjugate transpose of matrix is equal
            to the conjugate transpose of the derivative.
        """
        if hasattr(self, 'utry'):
            return np.array([])

        grads = self.gate.get_grad(params)  # type: ignore
        return np.transpose(grads.conj(), (0, 2, 1))

    def get_unitary_and_grad(
        self,
        params: RealVector = [],
    ) -> tuple[UnitaryMatrix, npt.NDArray[np.complex128]]:
        """
        Return the unitary and gradient for this gate.

        See :class:`DifferentiableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return self.utry, np.array([])

        utry, grads = self.gate.get_unitary_and_grad(params)  # type: ignore
        return utry.dagger, np.transpose(grads.conj(), (0, 2, 1))

    def optimize(self, env_matrix: npt.NDArray[np.complex128]) -> list[float]:
        """
        Return the optimal parameters with respect to an environment matrix.

        See :class:`LocallyOptimizableUnitary` for more info.
        """
        if hasattr(self, 'utry'):
            return []
        self.check_env_matrix(env_matrix)
        return self.gate.optimize(env_matrix.conj().T)  # type: ignore

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, DaggerGate)
            and self.gate == other.gate
        )

    def __hash__(self) -> int:
        return hash(self.gate)

    def get_inverse(self) -> Gate:
        """Return the gate's inverse as a gate."""
        return self.gate



"""This package implements BQSKit's gate retargeting passes."""
from __future__ import annotations

from bqskit.passes.retarget.auto import AutoRebase2QuditGatePass
from bqskit.passes.retarget.general import GeneralSQDecomposition
from bqskit.passes.retarget.two import Rebase2QuditGatePass

__all__ = [
    'AutoRebase2QuditGatePass',
    'GeneralSQDecomposition',
    'Rebase2QuditGatePass',
]



"""This module implements the RuntimeTask class."""
from __future__ import annotations

import inspect
import logging
from typing import Any
from typing import Coroutine

import dill

from bqskit.runtime.address import RuntimeAddress


class RuntimeTask:
    """
    BQSKit Runtime's internal task structure.

    The task exists in two phases:
        1. In the initial phase, the task is a context-less function
        and arguments grouped with a return address. It is expected
        when the task is finished executing, to ship the result to
        the return address.

        2. When the task has been scheduled to run on a worker, it is
        initialized into a coroutine that can be executed in steps.
    """

    task_counter = 0
    """Per-process task id counter."""

    def __init__(
        self,
        fnargs: tuple[Any, Any, Any],
        return_address: RuntimeAddress,
        comp_task_id: int,
        breadcrumbs: tuple[RuntimeAddress, ...],
        logging_level: int | None = None,
        max_logging_depth: int = -1,
        task_name: str | None = None,
        log_context: dict[str, str] = {},
    ) -> None:
        """Create the task with a new id and return address."""
        RuntimeTask.task_counter += 1
        self.task_id = RuntimeTask.task_counter

        self.serialized_fnargs = dill.dumps(fnargs)
        self._fnargs: tuple[Any, Any, Any] | None = None
        self._name = fnargs[0].__name__ if task_name is None else task_name
        """Tuple of function pointer, arguments, and keyword arguments."""

        self.return_address = return_address
        """
        Where the result of this task should be sent.

        This doubles as a unique system-wide id for the task.
        """

        self.logging_level = logging_level or 0
        """Logs with levels >= to this get emitted, if None always emit."""

        self.comp_task_id = comp_task_id
        """The mailbox id of the this task's root task."""

        self.breadcrumbs = breadcrumbs
        """All of this task's parent tasks' addresses in order."""

        self.max_logging_depth = max_logging_depth
        """Logs are not emitted for tasks with this many or more parents."""

        self.coro: Coroutine[Any, Any, Any] | None = None
        """The coroutine containing this tasks code."""

        self.desired_box_id: int | None = None
        """When waiting on a mailbox, this stores that mailbox's id."""

        self.owned_mailboxes: list[int] = []
        """The mailbox ids that this task owns."""

        self.wake_on_next: bool = False
        """Set to true if this task should wake immediately on a result."""

        self.log_context: dict[str, str] = log_context
        """Additional context to be logged with this task."""

        self.msg_buffer: list[Any] = []

    @property
    def fnargs(self) -> tuple[Any, Any, Any]:
        """Return the function pointer, arguments, and keyword arguments."""
        if self._fnargs is None:
            self._fnargs = dill.loads(self.serialized_fnargs)
        assert self._fnargs is not None  # for type checker
        return self._fnargs

    def step(self, send_val: Any = None) -> Any:
        """Execute one step of the task."""
        if self.coro is None:
            raise RuntimeError('Task has not been initialized.')

        # Reset previously set await flags
        self.wake_on_next = False
        self.desired_box_id = None

        # Set logging level
        old_level = logging.getLogger().getEffectiveLevel()
        if (
            self.max_logging_depth < 0
            or len(self.breadcrumbs) <= self.max_logging_depth
        ):
            logging.getLogger().setLevel(self.logging_level)
        else:
            logging.getLogger().setLevel(100)

        # Execute a task step
        to_return = self.coro.send(send_val)

        # Reset logging
        logging.getLogger().setLevel(old_level)

        return to_return

    @property
    def unique_id(self) -> RuntimeAddress:
        """Return the task's system-wide unique id."""
        return self.return_address

    def start(self) -> None:
        """Initialize the task."""
        self.coro = self.run()

    def cancel(self) -> None:
        """Ask the coroutine to gracefully exit."""
        if self.coro is not None:
            # If this call to "close" raises a RuntimeError,
            # it is likely a blanket try/accept catching the
            # error used to stop the coroutine, preventing
            # it from stopping correctly.
            try:
                self.coro.close()
            except ValueError:
                # Coroutine is running and cannot be closed.
                pass
        else:
            raise RuntimeError('Task was cancelled with None coroutine.')

    async def run(self) -> Any:
        """Task coroutine wrapper."""
        if inspect.iscoroutinefunction(self.fnargs[0]):
            return await self.fnargs[0](*self.fnargs[1], **self.fnargs[2])
        return self.fnargs[0](*self.fnargs[1], **self.fnargs[2])

    def is_descendant_of(self, addr: RuntimeAddress) -> bool:
        """Return true if `addr` identifies a parent (or this) task."""
        return addr == self.return_address or addr in self.breadcrumbs

    def __str__(self) -> str:
        """Return a string representation of the task."""
        return f'{self._name}'

    def __repr__(self) -> str:
        """Return a string representation of the task."""
        return f'<RuntimeTask {self.unique_id} {self._name}>'



"""This package implements predicates for use in control passes."""
from __future__ import annotations

from bqskit.passes.control.predicates.andpredicate import AndPredicate
from bqskit.passes.control.predicates.change import ChangePredicate
from bqskit.passes.control.predicates.count import GateCountPredicate
from bqskit.passes.control.predicates.many import ManyQuditGatesPredicate
from bqskit.passes.control.predicates.multi import MultiPhysicalPredicate
from bqskit.passes.control.predicates.notpredicate import NotPredicate
from bqskit.passes.control.predicates.orpredicate import OrPredicate
from bqskit.passes.control.predicates.physical import PhysicalPredicate
from bqskit.passes.control.predicates.single import AllConstantSingleQuditGates
from bqskit.passes.control.predicates.single import HasGeneralSingleQuditGate
from bqskit.passes.control.predicates.single import NoSingleQuditGatesInModel
from bqskit.passes.control.predicates.single import SinglePhysicalPredicate
from bqskit.passes.control.predicates.single import ZXGatePredicate
from bqskit.passes.control.predicates.width import WidthPredicate

__all__ = [
    'ChangePredicate',
    'GateCountPredicate',
    'AndPredicate',
    'NotPredicate',
    'OrPredicate',
    'WidthPredicate',
    'PhysicalPredicate',
    'SinglePhysicalPredicate',
    'MultiPhysicalPredicate',
    'ManyQuditGatesPredicate',
    'NoSingleQuditGatesInModel',
    'HasGeneralSingleQuditGate',
    'ZXGatePredicate',
    'AllConstantSingleQuditGates',
]



"""This module implements the CompilationStatus enum."""
from __future__ import annotations

from enum import IntEnum


class CompilationStatus(IntEnum):
    """The status of a CompilationTask."""
    UNKNOWN = 0
    RUNNING = 1
    DONE = 2



"""This module implements the StateVector class."""
from __future__ import annotations

import logging
import typing
from typing import Any
from typing import cast
from typing import Iterator
from typing import Sequence
from typing import TYPE_CHECKING
from typing import Union

import numpy as np
import numpy.typing as npt
from numpy.lib.mixins import NDArrayOperatorsMixin
from scipy.stats import unitary_group

from bqskit.utils.typing import is_integer
from bqskit.utils.typing import is_valid_radixes
from bqskit.utils.typing import is_vector

if TYPE_CHECKING:
    from typing import TypeGuard
    from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix
    from bqskit.ir.location import CircuitLocationLike

_logger = logging.getLogger(__name__)


class StateVector(NDArrayOperatorsMixin):
    """A vector representing a pure quantum state."""

    def __init__(
        self,
        input: StateLike,
        radixes: Sequence[int] = [],
        check_arguments: bool = True,
    ) -> None:
        """
        Constructs a `StateVector` from the supplied vector.

        Args:
            input (StateLike): The state vector input.

            radixes (Sequence[int]): A sequence with its length equal to
                the number of qudits this `StateVector` represents. Each
                element specifies the base, number of orthogonal states,
                for the corresponding qudit. By default, the constructor
                will attempt to calculate `radixes` from `input`.

            check_arguments (bool): If true, check arguments for type
                and value errors.

        Raises:
            ValueError: If `input` is not a pure quantum state.

            ValueError: If the dimension of `input` does not match the
                expected dimension from `radixes`.

            RuntimeError: If `radixes` is not specified and the
                constructor cannot infer it.
        """
        # Copy Constructor
        if isinstance(input, StateVector):
            self._vec = input.numpy
            self._radixes = input.radixes
            self._dim = input.dim
            return

        if check_arguments and not is_vector(input):
            raise TypeError(f'Expected vector, got {type(input)}.')

        if check_arguments and not StateVector.is_pure_state(input):
            raise ValueError('Input failed state vector condition.')

        dim = len(input)

        if radixes:
            self._radixes = tuple(radixes)

        # Check if unitary dimension is a power of two
        elif dim & (dim - 1) == 0:
            self._radixes = tuple([2] * int(np.round(np.log2(dim))))

        # Check if unitary dimension is a power of three
        elif 3 ** int(np.round(np.log(dim) / np.log(3))) == dim:  # noqa
            radixes = [3] * int(np.round(np.log(dim) / np.log(3)))
            self._radixes = tuple(radixes)

        else:
            raise RuntimeError(
                'Unable to determine radixes'
                ' for StateVector with dim %d.' % dim,
            )

        if check_arguments and not is_valid_radixes(self.radixes):
            raise TypeError('Invalid qudit radixes.')

        if check_arguments and np.prod(self.radixes) != dim:
            raise ValueError('Qudit radixes mismatch with dimension.')

        self._vec = np.array(input, dtype=np.complex128)
        self._dim = dim

    @property
    def numpy(self) -> npt.NDArray[np.complex128]:
        """The NumPy array holding the vector."""
        return self._vec

    @property
    def shape(self) -> tuple[int, ...]:
        """The one-dimensional shape of the vector."""
        return self._vec.shape

    @property
    def dtype(self) -> np.typing.DTypeLike:
        """The NumPy data type of the vector."""
        return self._vec.dtype

    @property
    def num_qudits(self) -> int:
        """The number of qudits in the state."""
        return len(self.radixes)

    @property
    def dim(self) -> int:
        """The vector dimension for this state."""
        return self._dim

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return self._radixes

    def __len__(self) -> int:
        """The dimension of the state vector."""
        return self.shape[0]

    def __iter__(self) -> Iterator[np.complex128]:
        """An iterator that iterates through the elements of the vector."""
        return self._vec.__iter__()

    def __getitem__(
            self, index: Any,
    ) -> np.complex128 | npt.NDArray[np.complex128]:
        """Implements NumPy API for the StateVector class."""
        return self._vec[index]

    def get_probs(self) -> tuple[float, ...]:
        """Return the probabilities for each classical outcome."""
        return tuple(np.abs(elem)**2 for elem in self)

    def is_qubit_only(self) -> bool:
        """Return true if this unitary can only act on qubits."""
        return all([radix == 2 for radix in self.radixes])

    def is_qutrit_only(self) -> bool:
        """Return true if this unitary can only act on qutrits."""
        return all([radix == 3 for radix in self.radixes])

    def is_qudit_only(self, radix: int) -> bool:
        """
        Return true if this unitary can only act on `radix`-qudits.

        Args:
            radix (int): Check all qudits have this many orthogonal
                states.
        """
        return all([r == radix for r in self.radixes])

    @staticmethod
    def is_pure_state(V: Any, tol: float = 1e-8) -> TypeGuard[StateLike]:
        """
        Check if V is a pure state vector.

        Args:
            V (Any): The vector to check.

            tol (float): The numerical precision of the check.

        Returns:
            bool: True if V is a pure quantum state vector.
        """
        if isinstance(V, StateVector):
            return True

        from bqskit.qis.state import StateSystem
        if isinstance(V, StateSystem):
            return False

        if not np.allclose(np.sum(np.square(np.abs(V))), 1, rtol=0, atol=tol):
            _logger.debug('Failed pure state criteria.')
            return False

        return True

    @staticmethod
    def zero(num_qudits: int, radixes: Sequence[int] = []) -> StateVector:
        """Prepares the zero state."""
        if len(radixes) == 0:
            radixes = [2] * num_qudits
        state = np.zeros(np.prod(radixes), dtype=np.complex128)
        state[0] = 1.0
        return StateVector(state, radixes)

    @staticmethod
    def random(num_qudits: int, radixes: Sequence[int] = []) -> StateVector:
        """
        Sample a random pure state.

        Args:
            num_qudits (np.ndarray): The number of qudits in the state.
                This is not the dimension.

            radixes (Sequence[int]): The radixes for the StateVector.

        Returns:
            StateVector: A random pue quantum state.

        Raises:
            ValueError: If `num_qudits` is nonpositive.

            ValueError: If the length of `radixes` is not equal to
                `num_qudits`.
        """
        if not is_integer(num_qudits):
            raise TypeError(
                f'Expected int for num_qudits, got {type(num_qudits)}.',
            )

        if num_qudits <= 0:
            raise ValueError('Expected positive number for num_qudits.')

        radixes = tuple(radixes if len(radixes) > 0 else [2] * num_qudits)

        if not is_valid_radixes(radixes):
            raise TypeError('Invalid qudit radixes.')

        if len(radixes) != num_qudits:
            raise ValueError(
                'Expected length of radixes to be equal to num_qudits:'
                ' %d != %d' % (len(radixes), num_qudits),
            )

        U = unitary_group.rvs(int(np.prod(radixes)))
        return StateVector(U[:, 0], radixes)

    def __hash__(self) -> int:
        """Hash the state vector."""
        return hash(tuple(self.numpy))

    def __eq__(self, other: object) -> bool:
        """Check if `self` is approximately equal to `other`."""
        if isinstance(other, StateVector):
            return np.allclose(self.numpy, other.numpy)

        if isinstance(other, np.ndarray):
            return np.allclose(self.numpy, other)

        return NotImplemented

    def apply(
        self,
        utry: UnitaryMatrix,
        location: CircuitLocationLike,
        inverse: bool = False,
        check_arguments: bool = True,
    ) -> None:
        """

        Apply the specified unitary on the right of this StateVector.

        ..
                .---.   .------.
                |   |---|      |- 0
                |   |---| utry |- 1
                .   .   '------'
                .   .
                .   .
                |   |------------ n-1
                '---'

        Args:
            utry (UnitaryMatrix): The unitary to apply.

            location (CircuitLocationLike): The qudits to apply the unitary on.

            inverse (bool): If true, apply the inverse of the unitary.

            check_arguments (bool): If true, check the inputs for type and
                value errors.

        Raises:
            ValueError: If `utry`'s size does not match the given location.

            ValueError: if `utry`'s radixes does not match the given location.

        Notes:
            - Applying the unitary on the left is equivalent to multiplying
              the unitary on the right of the tensor. The notation comes
              from the quantum circuit perspective.

            - This operation is performed using tensor contraction.
        """
        from bqskit.ir.location import CircuitLocation
        from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix

        if check_arguments:
            if not isinstance(utry, UnitaryMatrix):
                raise TypeError('Expected UnitaryMatrix, got %s', type(utry))

            if not CircuitLocation.is_location(location, self.num_qudits):
                raise TypeError('Invalid location.')

            location = CircuitLocation(location)

            if len(location) != utry.num_qudits:
                raise ValueError('Unitary and location size mismatch.')

            for utry_radix, bldr_radix_idx in zip(utry.radixes, location):
                if utry_radix != self.radixes[bldr_radix_idx]:
                    raise ValueError('Unitary and location radix mismatch.')

        location = cast(CircuitLocation, location)
        qudits = list(range(self.num_qudits))
        identity_action_perm = [
            x
            for x in qudits
            if x not in location
        ]
        unitary_action_perm = list(location)

        left_dim = int(
            np.prod([
                self.radixes[x]
                for x in unitary_action_perm
            ]),
        )

        utry = utry.dagger if inverse else utry

        perm = unitary_action_perm + identity_action_perm
        self._vec = self._vec.reshape(self.radixes)
        self._vec = self._vec.transpose(perm)
        self._vec = self._vec.reshape((left_dim, -1))
        self._vec = utry @ self._vec

        shape = list(self.radixes) * 2
        shape = [shape[p] for p in perm]
        self._vec = self._vec.reshape(shape)
        inv_perm = list(np.argsort(perm))
        self._vec = self._vec.transpose(inv_perm)
        self._vec = self._vec.reshape(-1)

    def get_distance_from(self, other: StateLike) -> float:
        """
        Return the distance between `self` and `other`.

        The distance is given as the infidelity between the two states.

        Args:
            other (StateLike): The state to measure distance from.

        Returns:
            float: A value between 1 and 0, where 0 means the two states
            are equal up to global phase and 1 means the two states are
            very unsimilar or far apart.
        """
        other = StateVector(other)
        dist = 1 - np.abs(np.conj(self) @ other) ** 2
        return typing.cast(float, dist if dist > 0.0 else 0.0)

    def __array__(
        self,
        dtype: np.typing.DTypeLike = np.complex128,
    ) -> npt.NDArray[np.complex128]:
        """Implements NumPy API for the StateVector class."""
        if dtype != np.complex128:
            raise ValueError('StateVector only supports Complex128 dtype.')

        return self._vec

    def __array_ufunc__(
        self,
        ufunc: np.ufunc,
        method: str,
        *inputs: npt.NDArray[Any],
        **kwargs: Any,
    ) -> StateVector | npt.NDArray[np.complex128]:
        """Implements NumPy API for the StateVector class."""
        if method != '__call__':
            return NotImplemented

        non_state_involved = False
        args: list[npt.NDArray[Any]] = []
        for input in inputs:
            if isinstance(input, StateVector):
                args.append(input.numpy)
            else:
                args.append(input)
                non_state_involved = True

        out = ufunc(*args, **kwargs)

        # The results are state vectors
        # if only states are involved
        # and state vectors are closed under the specific operation.
        convert_back = (
            not non_state_involved and ufunc.__name__ == 'conjugate'
            or (
                ufunc.__name__ == 'multiply'
                and all(
                    np.isscalar(input) or isinstance(input, StateVector)
                    for input in inputs
                )
                and all(
                    np.abs(np.abs(input) - 1) <= 1e-14
                    for input in inputs if np.isscalar(input)
                )
            )
        )

        if convert_back:
            return StateVector(out, self.radixes)

        return out

    def __str__(self) -> str:
        """Return the string representation of the vector."""
        return str(self._vec)

    def __repr__(self) -> str:
        """Return the repr representation of the vector."""
        return repr(self._vec)


StateLike = Union[
    StateVector,
    npt.NDArray[np.complex128],
    Sequence[Union[int, float, complex]],
]



"""This module implements the SqrtCNOTGate."""
from __future__ import annotations

from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.qis.unitary.unitarymatrix import UnitaryMatrix


class SqrtCNOTGate(ConstantGate, QubitGate):
    """
    The Square root Controlled-X gate.

    The SqrtCNOT gate is given by the following unitary:

    .. math::

        \\begin{pmatrix}
        1 & 0 & 0 & 0 \\\\
        0 & 1 & 0 & 0 \\\\
        0 & 0 & \\frac{1}{2} + \\frac{1}{2}i & \\frac{1}{2} - \\frac{1}{2}i \\\\
        0 & 0 & \\frac{1}{2} - \\frac{1}{2}i & \\frac{1}{2} + \\frac{1}{2}i \\\\
        \\end{pmatrix}
    """

    _num_qudits = 2
    _qasm_name = 'csx'
    _utry = UnitaryMatrix(
        [
            [1, 0, 0, 0],
            [0, 1, 0, 0],
            [0, 0, 0.5 + 0.5j, 0.5 - 0.5j],
            [0, 0, 0.5 - 0.5j, 0.5 + 0.5j],
        ],
    )



"""This package implements language reading and writing features."""
from __future__ import annotations

from bqskit.ir.lang.language import Language

__all__ = ['Language']

_language_dict: dict[str, Language] = {}


def register_language(extension: str, language: Language) -> None:
    _language_dict[extension] = language


def get_language(extension: str) -> Language:
    if extension not in _language_dict:
        raise ValueError(f'Unsupported extension: {extension}.')

    return _language_dict[extension]



"""This module implements the ScanningGateRemovalPass."""
from __future__ import annotations

import logging
from typing import Any
from typing import Callable

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.operation import Operation
from bqskit.ir.opt.cost.functions import HilbertSchmidtResidualsGenerator
from bqskit.ir.opt.cost.generator import CostFunctionGenerator
from bqskit.utils.typing import is_real_number
_logger = logging.getLogger(__name__)


class ScanningGateRemovalPass(BasePass):
    """
    The ScanningGateRemovalPass class.

    Starting from one side of the circuit, attempt to remove gates one-by-one.
    """

    def __init__(
        self,
        start_from_left: bool = True,
        success_threshold: float = 1e-8,
        cost: CostFunctionGenerator = HilbertSchmidtResidualsGenerator(),
        instantiate_options: dict[str, Any] = {},
        collection_filter: Callable[[Operation], bool] | None = None,
    ) -> None:
        """
        Construct a ScanningGateRemovalPass.

        Args:
            start_from_left (bool): Determines where the scan starts
                attempting to remove gates from. If True, scan goes left
                to right, otherwise right to left. (Default: True)

            success_threshold (float): The distance threshold that
                determines successful termintation. Measured in cost
                described by the hilbert schmidt cost function.
                (Default: 1e-8)

            cost (CostFunction | None): The cost function that determines
                successful removal of a gate.
                (Default: HilbertSchmidtResidualsGenerator())

            instantiate_options (dict[str: Any]): Options passed directly
                to circuit.instantiate when instantiating circuit
                templates. (Default: {})

            collection_filter (Callable[[Operation], bool] | None):
                A predicate that determines which operations should be
                attempted to be removed. Called with each operation
                in the circuit. If this returns true, this pass will
                attempt to remove that operation. Defaults to all
                operations.
        """

        if not is_real_number(success_threshold):
            raise TypeError(
                'Expected real number for success_threshold'
                ', got %s' % type(success_threshold),
            )

        if not isinstance(cost, CostFunctionGenerator):
            raise TypeError(
                'Expected cost to be a CostFunctionGenerator, got %s'
                % type(cost),
            )

        if not isinstance(instantiate_options, dict):
            raise TypeError(
                'Expected dictionary for instantiate_options, got %s.'
                % type(instantiate_options),
            )

        self.collection_filter = collection_filter or default_collection_filter

        if not callable(self.collection_filter):
            raise TypeError(
                'Expected callable method that maps Operations to booleans for'
                ' collection_filter, got %s.' % type(self.collection_filter),
            )

        self.start_from_left = start_from_left
        self.success_threshold = success_threshold
        self.cost = cost
        self.instantiate_options: dict[str, Any] = {
            'dist_tol': self.success_threshold,
            'min_iters': 100,
            'cost_fn_gen': self.cost,
        }
        self.instantiate_options.update(instantiate_options)

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        instantiate_options = self.instantiate_options.copy()
        if 'seed' not in instantiate_options:
            instantiate_options['seed'] = data.seed

        start = 'left' if self.start_from_left else 'right'
        _logger.debug(f'Starting scanning gate removal on the {start}.')

        target = self.get_target(circuit, data)

        circuit_copy = circuit.copy()
        reverse_iter = not self.start_from_left
        for cycle, op in circuit.operations_with_cycles(reverse=reverse_iter):

            if not self.collection_filter(op):
                _logger.debug(f'Skipping operation {op} at cycle {cycle}.')
                continue

            _logger.debug(f'Attempting removal of operation at cycle {cycle}.')
            _logger.debug(f'Operation: {op}')

            working_copy = circuit_copy.copy()

            # If removing gates from the left, we need to track index changes.
            if self.start_from_left:
                idx_shift = circuit.num_cycles
                idx_shift -= working_copy.num_cycles
                cycle -= idx_shift

            working_copy.pop((cycle, op.location[0]))
            working_copy.instantiate(target, **instantiate_options)

            if self.cost(working_copy, target) < self.success_threshold:
                _logger.debug('Successfully removed operation.')
                circuit_copy = working_copy

        circuit.become(circuit_copy)


def default_collection_filter(op: Operation) -> bool:
    return True



"""
BQSKit Gates (:mod:`bqskit.ir.gates`)
=====================================

.. rubric:: Constant Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    BGate
    CCXGate
    ToffoliGate
    CHGate
    ClockGate
    CPIGate
    CSGate
    CSUMGate
    CTGate
    CNOTGate
    CXGate
    CYGate
    CZGate
    ECRGate
    HGate
    IdentityGate
    ISwapGate
    IToffoliGate
    PDGate
    PermutationGate
    MargolusGate
    RC3XGate
    RCCXGate
    SGate
    SdgGate
    ShiftGate
    SqrtCNOTGate
    SqrtISwapGate
    SqrtTGate
    SubSwapGate
    SwapGate
    SqrtXGate
    SXGate
    SycamoreGate
    TGate
    TdgGate
    ConstantUnitaryGate
    XGate
    XXGate
    YGate
    YYGate
    ZGate
    ZZGate

.. rubric:: Parameterized Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    CCPGate
    CKMGate
    CKMdgGate
    CPGate
    ArbitraryCPhaseGate
    CRXGate
    CRYGate
    CRZGate
    CUGate
    FSIMGate
    MPRYGate
    MPRZGate
    PauliGate
    PauliZGate
    PhasedXZGate
    RSU3Gate
    RXGate
    RXXGate
    RYGate
    RYYGate
    RZGate
    RZZGate
    U1Gate
    U1qGate
    U1qPi2Gate
    U1qPiGate
    U2Gate
    U3Gate
    U8Gate
    VariableUnitaryGate

.. rubric:: Composed Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    ControlledGate
    PowerGate
    DaggerGate
    EmbeddedGate
    FrozenParameterGate
    TaggedGate
    VariableLocationGate

.. rubric:: Special Gates

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:

    CircuitGate
    MeasurementPlaceholder
    Reset
    BarrierPlaceholder

.. rubric:: Gate Base Classes

.. autosummary::
    :toctree: autogen
    :recursive:
    :nosignatures:
    :template: autosummary/gate.rst

    ComposedGate
    QubitGate
    QutritGate
    QuditGate
    ConstantGate
    GeneralGate
"""
from __future__ import annotations

from bqskit.ir.gates.barrier import BarrierPlaceholder
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.composed import *  # noqa
from bqskit.ir.gates.composed import __all__ as composed_all
from bqskit.ir.gates.composedgate import ComposedGate
from bqskit.ir.gates.constant import *  # noqa
from bqskit.ir.gates.constant import __all__ as constant_all
from bqskit.ir.gates.constantgate import ConstantGate
from bqskit.ir.gates.generalgate import GeneralGate
from bqskit.ir.gates.measure import MeasurementPlaceholder
from bqskit.ir.gates.parameterized import *  # noqa
from bqskit.ir.gates.parameterized import __all__ as parameterized_all
from bqskit.ir.gates.qubitgate import QubitGate
from bqskit.ir.gates.quditgate import QuditGate
from bqskit.ir.gates.qutritgate import QutritGate
from bqskit.ir.gates.reset import Reset

__all__ = composed_all + constant_all + parameterized_all
__all__ += ['ComposedGate', 'ConstantGate']
__all__ += ['QubitGate', 'QutritGate', 'QuditGate']
__all__ += [
    'CircuitGate', 'MeasurementPlaceholder',
    'Reset', 'BarrierPlaceholder',
]
__all__ += ['GeneralGate']

# TODO: Implement the rest of the gates in:
# https://pubs.aip.org/aip/jmp/article-abstract/56/3/032202/763827

# TODO: Implement generalization of CZ and CZD
# https://arxiv.org/abs/2206.07216



"""This module implements functions for translating to and from QuTiP."""
from __future__ import annotations

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from qutip import QubitCircuit

from bqskit.ir.circuit import Circuit
from bqskit.ir.lang.qasm2 import OPENQASM2Language


def qutip_to_bqskit(qc: QubitCircuit) -> Circuit:
    """Convert QuTiP's QubitCircuit `qc` to a BQSKit Circuit."""
    try:
        from qutip.qip.qasm import circuit_to_qasm_str
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import qutip package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install qutip\n',
        ) from e

    circuit = OPENQASM2Language().decode(circuit_to_qasm_str(qc))
    return circuit


def bqskit_to_qutip(circuit: Circuit) -> QubitCircuit:
    """Convert a BQSKit Circuit to QuTiP's QubitCircuit."""
    try:
        from qutip.qip.qasm import read_qasm
    except ImportError as e:
        raise ImportError(
            '\n\nUnable to import qutip package.\n'
            'Please ensure that it is installed with the following command:\n'
            '\tpip install qutip\n',
        ) from e

    return read_qasm(OPENQASM2Language().encode(circuit), strmode=True)



"""
=============================================================
Circuit Intermediate Representation (:mod:`bqskit.ir`)
=============================================================

.. currentmodule:: bqskit.ir

The BQSKit Circuit structure is a 2-d array of operations. A circuit is
indexed first by cycles and second by qudits. The cycle index determines
when operations are executed, and the qudit index determines which
qudits an operation operates on.

Every component of the IR is also a function from a vector of
real numbers to a unitary matrix. This is to facilitate circuit
instantiation, a core primitive in quantum synthesis.

.. rubric:: Core Classes

.. autosummary::
    :toctree: autogen
    :recursive:

    Circuit
    Gate
    Operation

.. rubric:: Circuit Indexing and Helpers

.. autosummary::
    :toctree: autogen
    :recursive:

    CycleInterval
    CircuitIterator
    CircuitLocation
    CircuitPoint
    CircuitRegion
    CircuitStructure

.. rubric:: Type Aliases

.. autosummary::
    :toctree: autogen
    :recursive:

    IntervalLike
    CircuitLocationLike
    CircuitPointLike
    CircuitRegionLike

.. rubric:: Gate Library

.. automodule:: bqskit.ir.gates
   :no-members:
   :no-inherited-members:
   :no-special-members:
"""
from __future__ import annotations

from bqskit.ir.circuit import Circuit
from bqskit.ir.gate import Gate
from bqskit.ir.interval import CycleInterval
from bqskit.ir.interval import IntervalLike
from bqskit.ir.iterator import CircuitIterator
from bqskit.ir.lang import register_language as _register_language
from bqskit.ir.lang.qasm2 import OPENQASM2Language as _qasm
from bqskit.ir.location import CircuitLocation
from bqskit.ir.location import CircuitLocationLike
from bqskit.ir.operation import Operation
from bqskit.ir.point import CircuitPoint
from bqskit.ir.point import CircuitPointLike
from bqskit.ir.region import CircuitRegion
from bqskit.ir.region import CircuitRegionLike
from bqskit.ir.structure import CircuitStructure


# Register supported languages
_register_language('qasm', _qasm())


__all__ = [
    'Operation',
    'Circuit',
    'Gate',
    'CircuitIterator',
    'CircuitLocation',
    'CircuitLocationLike',
    'CircuitPoint',
    'CircuitPointLike',
    'CircuitRegion',
    'CircuitRegionLike',
    'CycleInterval',
    'IntervalLike',
    'CircuitStructure',
]



"""This module implements the BlockConversionPass."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.gates.circuitgate import CircuitGate
from bqskit.ir.gates.constant.unitary import ConstantUnitaryGate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate
from bqskit.ir.point import CircuitPoint

_logger = logging.getLogger(__name__)


class BlockConversionPass(BasePass):
    """
    Converts blocks of one type to another type.

    Blocks are either described by a constant or variable unitary gate or as a
    circuit gate. Often during the flow of compilatin we will need to convert
    them from one form to another for a future pass.
    """

    def __init__(
        self,
        convert_target: str,
        convert_variable: bool = True,
        convert_constant: bool = True,
        convert_circuitgates: bool = True,
    ):
        """
        Construct a BlockConversionPass.

        Args:
            convert_target (str): Either `variable` or `constant`.
                Blocks will be converted to the form described here. If
                this is `variable` all gates will be converted to
                `VariableUnitaryGate` s. If this is `constant` blocks
                will be converted to `ConstantUnitaryGate` s. Blocks
                cannot be converted to circuit gates, that can be caried
                out by synthesis.

            convert_variable (bool): If this is true, this will replace
                VariableUnitaryGate's in the circuit with one's specified
                in convert_target.

            convert_constant (bool): If this is true, this will replace
                ConstantUnitaryGate's in the circuit with one's specified
                in convert_target.

            convert_circuitgates (bool): If this is true, this will replace
                CircuitGate's in the circuit with one's specified
                in convert_target. The subcircuit information captured
                in the circuit gate will be lost.
        """
        self.convert_variable = convert_variable
        self.convert_constant = convert_constant
        self.convert_circuitgates = convert_circuitgates

        if convert_target == 'variable':
            self.convert_target = 'variable'
            self.convert_variable = False
        elif convert_target == 'constant':
            self.convert_target = 'constant'
            self.convert_constant = False
        else:
            raise ValueError('Unexpected input for conversion target.')

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""

        # Variable -> Constant
        if self.convert_variable and self.convert_target == 'constant':
            _logger.debug('Converting variable gates to constant gates.')

            for cycle, op in circuit.operations_with_cycles():
                if isinstance(op.gate, VariableUnitaryGate):
                    cgate = ConstantUnitaryGate(op.get_unitary(), op.radixes)
                    point = CircuitPoint(cycle, op.location[0])
                    circuit.replace_gate(point, cgate, op.location)

        # CircuitGates -> Constant
        if self.convert_circuitgates and self.convert_target == 'constant':
            _logger.debug('Converting circuit gates to constant gates.')

            for cycle, op in circuit.operations_with_cycles():
                if isinstance(op.gate, CircuitGate):
                    cgate = ConstantUnitaryGate(op.get_unitary(), op.radixes)
                    point = CircuitPoint(cycle, op.location[0])
                    circuit.replace_gate(point, cgate, op.location)

        # Constant -> Variable
        if self.convert_constant and self.convert_target == 'variable':
            _logger.debug('Converting constant gates to variable gates.')

            for cycle, op in circuit.operations_with_cycles():
                if isinstance(op.gate, ConstantUnitaryGate):
                    params = VariableUnitaryGate.get_params(op.get_unitary())
                    vgate = VariableUnitaryGate(op.num_qudits, op.radixes)
                    point = CircuitPoint(cycle, op.location[0])
                    circuit.replace_gate(point, vgate, op.location, params)

        # CircuitGates -> Variable
        if self.convert_constant and self.convert_target == 'variable':
            _logger.debug('Converting circuit gates to variable gates.')

            for cycle, op in circuit.operations_with_cycles():
                if isinstance(op.gate, CircuitGate):
                    params = VariableUnitaryGate.get_params(op.get_unitary())
                    vgate = VariableUnitaryGate(op.num_qudits, op.radixes)
                    point = CircuitPoint(cycle, op.location[0])
                    circuit.replace_gate(point, vgate, op.location, params)



"""This module implements the StateSystem class."""
from __future__ import annotations

from typing import Any
from typing import Dict
from typing import Iterator
from typing import Mapping
from typing import TYPE_CHECKING
from typing import Union

import numpy as np

from bqskit.qis.state.state import StateVector

if TYPE_CHECKING:
    import numpy.typing as npt
    from typing import TypeGuard


class StateSystem(Mapping[StateVector, StateVector]):
    """A system of input and output states."""

    def __init__(self, system: StateSystemLike) -> None:
        """Construct a state system."""
        if isinstance(system, StateSystem):
            self._system: dict[StateVector, StateVector] = system._system
            self._radixes: tuple[int, ...] = system._radixes
            self._dim: int = system._dim
            self._vec_count: int = system._vec_count
            self.target: npt.NDArray[np.complex128] = system.target
            return

        self._system = {
            StateVector(k): StateVector(v)
            for k, v in system.items()
        }
        self._radixes = list(self._system.keys())[0].radixes
        self._dim = list(self._system.keys())[0].dim
        self._vec_count = len(self._system)

        for k, v in self._system.items():
            if k.radixes != self.radixes:
                raise ValueError('States in system have mismatch in dimension.')
            if v.radixes != self.radixes:
                raise ValueError('States in system have mismatch in dimension.')

        # Check overlap matrices
        V = np.column_stack(list(self._system.keys()))
        W = np.column_stack(list(self._system.values()))
        Ov = V.conj().T @ V
        Wv = W.conj().T @ W
        if not np.allclose(Ov, Wv):
            raise ValueError(
                'State system is unsolvable:'
                ' input and output overlap matrices do not match.',
            )
        self.target = W @ V.conj().T

    @property
    def num_qudits(self) -> int:
        """The number of qudits in the state."""
        return len(self.radixes)

    @property
    def dim(self) -> int:
        """The vector dimension for this state."""
        return self._dim

    @property
    def radixes(self) -> tuple[int, ...]:
        """The number of orthogonal states for each qudit."""
        return self._radixes

    def __iter__(self) -> Iterator[StateVector]:
        return self._system.__iter__()

    def __len__(self) -> int:
        return self._system.__len__()

    def __getitem__(self, _key: StateVector) -> StateVector:
        return self._system.__getitem__(_key)

    def __contains__(self, _key: object) -> bool:
        return self._system.__contains__(_key)

    def is_qubit_only(self) -> bool:
        """Return true if this unitary can only act on qubits."""
        return all([radix == 2 for radix in self.radixes])

    def is_qutrit_only(self) -> bool:
        """Return true if this unitary can only act on qutrits."""
        return all([radix == 3 for radix in self.radixes])

    @staticmethod
    def is_state_system(V: Any) -> TypeGuard[StateSystemLike]:
        """
        Check if V is a system of pure states.

        Args:
            V (Any): The vector to check.

        Returns:
            bool: True if V is a system of pure states.
        """
        if isinstance(V, StateSystem):
            return True

        if not isinstance(V, dict):
            return False

        for k, v in V.items():
            if not StateVector.is_pure_state(k):
                return False

            if not StateVector.is_pure_state(v):
                return False

        return True


StateSystemLike = Union[StateSystem, Dict[Any, Any]]



"""This module implements the FillSingleQuditGatesPass class."""
from __future__ import annotations

import logging

from bqskit.compiler.basepass import BasePass
from bqskit.compiler.passdata import PassData
from bqskit.ir.circuit import Circuit
from bqskit.ir.point import CircuitPoint


_logger = logging.getLogger(__name__)


class FillSingleQuditGatesPass(BasePass):
    """
    A pass that inserts single-qudit gates around multi-qudit gates.

    This pass will preserve the multi-qudit gates in the circuit and then place
    single-qudit unitary gates around those gates.
    """

    async def run(self, circuit: Circuit, data: PassData) -> None:
        """Perform the pass's operation, see :class:`BasePass` for more."""
        _logger.debug('Filling circuit with single-qudit gates.')

        complete_circuit = Circuit(circuit.num_qudits, circuit.radixes)
        sq_gate = data.gate_set.get_general_sq_gate()
        id_params = sq_gate.identity_as_params((circuit.radixes[0],))

        # Add general gate as identity on every qudit to start
        for qudit_index in range(circuit.num_qudits):
            complete_circuit.append_gate(sq_gate, qudit_index, id_params)

        for cycle, op in circuit.operations_with_cycles():

            # Single-qudit gates get converted to general gates
            if op.num_qudits == 1:

                # Check for already existing single-qudit gates
                last_point = complete_circuit.last_on(op.location[0])
                if last_point is not None:
                    last_op = complete_circuit[last_point]

                    if last_op.num_qudits == 1:
                        # Merge single-qudit gates if one already exists
                        utry = op.get_unitary() @ last_op.get_unitary()
                        last_op.params = sq_gate.calc_params(utry)
                        continue

                # Otherwise just add general gate
                params = sq_gate.calc_params(op.get_unitary())
                complete_circuit.append_gate(sq_gate, op.location, params)

            else:
                # Multi-qudit gates get added as is
                complete_circuit.append(op)

                # Add additional sq gates where there is none between mq gates
                qudits_needing_single_qudit_gates = set(op.location)
                for p in circuit.next(CircuitPoint(cycle, op.location[0])):
                    if circuit[p].num_qudits == 1:
                        qudits_needing_single_qudit_gates.remove(p.qudit)

                for qudit in qudits_needing_single_qudit_gates:
                    complete_circuit.append_gate(sq_gate, qudit, id_params)

        circuit.become(complete_circuit)



"""This package contains parameterized gates."""
from __future__ import annotations

from bqskit.ir.gates.parameterized.ccp import CCPGate
from bqskit.ir.gates.parameterized.ckm import CKMGate
from bqskit.ir.gates.parameterized.ckmd import CKMdgGate
from bqskit.ir.gates.parameterized.cp import CPGate
from bqskit.ir.gates.parameterized.cphase import ArbitraryCPhaseGate
from bqskit.ir.gates.parameterized.crx import CRXGate
from bqskit.ir.gates.parameterized.cry import CRYGate
from bqskit.ir.gates.parameterized.crz import CRZGate
from bqskit.ir.gates.parameterized.cu import CUGate
from bqskit.ir.gates.parameterized.diagonal import DiagonalGate
from bqskit.ir.gates.parameterized.fsim import FSIMGate
from bqskit.ir.gates.parameterized.mpry import MPRYGate
from bqskit.ir.gates.parameterized.mprz import MPRZGate
from bqskit.ir.gates.parameterized.pauli import PauliGate
from bqskit.ir.gates.parameterized.pauliz import PauliZGate
from bqskit.ir.gates.parameterized.phasedxz import PhasedXZGate
from bqskit.ir.gates.parameterized.rsu3 import RSU3Gate
from bqskit.ir.gates.parameterized.rx import RXGate
from bqskit.ir.gates.parameterized.rxx import RXXGate
from bqskit.ir.gates.parameterized.ry import RYGate
from bqskit.ir.gates.parameterized.ryy import RYYGate
from bqskit.ir.gates.parameterized.rz import RZGate
from bqskit.ir.gates.parameterized.rzz import RZZGate
from bqskit.ir.gates.parameterized.u1 import U1Gate
from bqskit.ir.gates.parameterized.u1q import U1qGate
from bqskit.ir.gates.parameterized.u1q import U1qPi2Gate
from bqskit.ir.gates.parameterized.u1q import U1qPiGate
from bqskit.ir.gates.parameterized.u2 import U2Gate
from bqskit.ir.gates.parameterized.u3 import U3Gate
from bqskit.ir.gates.parameterized.u8 import U8Gate
from bqskit.ir.gates.parameterized.unitary import VariableUnitaryGate

__all__ = [
    'CCPGate',
    'CKMGate',
    'CKMdgGate',
    'CPGate',
    'ArbitraryCPhaseGate',
    'CRXGate',
    'CRYGate',
    'CRZGate',
    'CUGate',
    'DiagonalGate',
    'FSIMGate',
    'MPRYGate',
    'MPRZGate',
    'PauliGate',
    'PauliZGate',
    'PhasedXZGate',
    'RSU3Gate',
    'RXGate',
    'RXXGate',
    'RYGate',
    'RYYGate',
    'RZGate',
    'RZZGate',
    'U1Gate',
    'U1qGate',
    'U1qPi2Gate',
    'U1qPiGate',
    'U2Gate',
    'U3Gate',
    'U8Gate',
    'VariableUnitaryGate',
]
